/***********************************************************************************
 *                 BatchWizardMod for VirtualDub and VirtualDubMod                 *
 *                                                                                 *
 * Copyright 2013 by Shengalts Aleksander aka Instructor (Shengalts@mail.ru)       *
 *                                                                                 *
 * License: this source is distributed under "BSD license" conditions.             *
 ***********************************************************************************/

#define WIN32_LEAN_AND_MEAN
#define _WIN32_IE 0x0500
#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include "StrFunc.h"
#include "WideFunc.h"
#include "StackFunc.h"
#include "RegExpFunc.h"
#include "Resources\resource.h"

//Include string functions
#define WideCharUpper
#define WideCharLower
#define xarraysizeA
#define xarraysizeW
#define xmemcpy
#define xmemset
#define xstrlenW
#define xstrcpyW
#define xstrcpynW
#define xstrcmpnW
#define xstrcmpiW
#define xstrcmpinW
#define xstrstrW
#define xstrrepW
#define xatoiW
#define xuitoaW
#define xitoaW
#define hex2decW
#define dec2hexW
#define xprintfW
#include "StrFunc.h"

//Include wide functions
#define AppendMenuWide
#define ComboBox_FindStringExactWide
#define ComboBox_InsertStringWide
#define CreateDialogWide
#define CreateDirectoryWide
#define CreateFileWide
#define CreateProcessWide
#define DialogBoxWide
#define DispatchMessageWide
#define DragQueryFileWide
#define GetFileAttributesWide
#define GetMessageWide
#define GetModuleFileNameWide
#define GetOpenFileNameWide
#define GetSaveFileNameWide
#define GetTempFileNameWide
#define GetTempPathWide
#define GetWindowLongPtrWide
#define GetWindowTextLengthWide
#define GetWindowTextWide
#define IsDialogMessageWide
#define ListView_GetItemWide
#define ListView_InsertItemWide
#define ListView_SetItemWide
#define SetDlgItemTextWide
#define SetWindowTextWide
#define SHBrowseForFolderWide
#define SHGetPathFromIDListWide
#include "WideFunc.h"

//Include stack functions
#define StackInsertBefore
#define StackDelete
#define StackJoin
#include "StackFunc.h"

//Include RegExp functions
#define RE_FUNCTIONS
#include "RegExpFunc.h"

//Defines
#define STRID_HELP                      1
#define STRID_OUTPUTDIRNOTFOUND         2
#define STRID_OUTPUTDIREMPTY            3
#define STRID_OUTPUTDIRNOTFOUNDCONTINUE 4
#define STRID_VIRTUALDUBNOTFOUND        5
#define STRID_MULTIFILESFIRST           6
#define STRID_MULTIFILESLAST            7
#define STRID_MULTIPASS                 8
#define STRID_ADDOPENSAVECOMMANDS       9
#define STRID_NOINPUTFILE               10
#define STRID_REMOVESUBSET              11
#define STRID_NOFILESFORPROCESS         12
#define STRID_CANNOTCREATEFILE          13
#define STRID_CREATETEMPLATE            14
#define STRID_DONEFILES                 15
#define STRID_SCRIPTFILE                16
#define STRID_GENERATE                  17
#define STRID_LISTLABEL                 18
#define STRID_OUTPUTDIR                 19
#define STRID_OUTPUTFILTER              20
#define STRID_FINDIT                    21
#define STRID_REPLACEWITH               22
#define STRID_MATCHCASE                 23
#define STRID_REGEXP                    24
#define STRID_CREATE                    25
#define STRID_EXECUTE                   26
#define STRID_SAVE                      27
#define STRID_INPUTFILE                 28
#define STRID_OUTPUTFILENAME            29
#define STRID_MENU_ITEMDELETE           30
#define STRID_MENU_ITEMMOVEUP           31
#define STRID_MENU_ITEMMOVEDOWN         32
#define STRID_MENU_ITEMCLEAR            33
#define STRID_TEMPLATE                  34
#define STRID_SCRIPTTYPE                35
#define STRID_SCRIPTVIRTUALDUB          36
#define STRID_SCRIPTAVS                 37
#define STRID_NOPARSE                   38
#define STRID_EMPTY                     39
#define STRID_CLOSE                     40
#define STRID_OK                        41
#define STRID_CANCEL                    42

#define BUFFER_SIZE          1024
#define COMMANDLINE_SIZE     32768

#define LVSI_INPUTFILE       0
#define LVSI_OUTPUTFILENAME  1

#define STR_DIALOGTITLE  L"BatchWizardMod"

#define STYPE_NONE                  0x0
#define STYPE_VIRTUALDUB            0x1
#define STYPE_AVS                   0x2
#define STYPE_AVST                  0x4

#define DSTATE_NONE                 0x0
#define DSTATE_NOOUTPUTFILE         0x1
#define DSTATE_SINGLEFILESINGLEPASS 0x2
#define DSTATE_SINGLEFILEMULTIPASS  0x4
#define DSTATE_MULTIFILES           0x8

#ifndef BIF_NEWDIALOGSTYLE
  #define BIF_NEWDIALOGSTYLE 0x0040
#endif
#ifndef INVALID_FILE_ATTRIBUTES
  #define INVALID_FILE_ATTRIBUTES ((DWORD)-1)
#endif

//DialogResizeMessages flags
#define DRM_PAINTSIZEGRIP 0x2 //Draw resize grid.

//DIALOGRESIZE type
#define DRS_SIZE  0x1 //Resize control. Can be combined with DRS_X ot DRS_Y.
#define DRS_MOVE  0x2 //Move control. Can be combined with DRS_X ot DRS_Y.
#define DRS_X     0x4 //X value. Can be combined with DRS_SIZE ot DRS_MOVE.
#define DRS_Y     0x8 //Y value. Can be combined with DRS_SIZE ot DRS_MOVE.

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
  int nOffset;   //Control offset, set it to zero
} DIALOGRESIZE;

//Functions prototypes
LRESULT CALLBACK MainDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK TemplateDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
int CALLBACK BrowseCallbackProc(HWND hWnd, UINT uMsg, LPARAM lParam, LPARAM lpData);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcMinMax, RECT *rcCurrent, DWORD dwFlags, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL ListSelItem(HWND hWnd, int nIndex);
int ListSetItem(HWND hWnd, int nIndex, const wchar_t *wpFile, const wchar_t *wpFileName, BOOL bNew);
DWORD ReadAnsiFileAsUnicode(const wchar_t *wpFile, wchar_t **wpContent);
void ParseScript(HWND hDlg, wchar_t *wpInput, INT_PTR nInputLen, wchar_t **wppOutput, INT_PTR *lpnOutputLen, BOOL bSkipParse);
int SaveScript(HWND hDlg, HWND hWndList, const wchar_t *wpOutputScriptFile);
int TextReplace(wchar_t **wppText, INT_PTR *lpnTextLen, const wchar_t *wpIt, int nItLen, const wchar_t *wpWith, int nWithLen, BOOL bSensitive);
int TextReplaceRE(wchar_t **wppText, INT_PTR *lpnTextLen, const wchar_t *wpIt, int nItLen, const wchar_t *wpWith, int nWithLen, BOOL bSensitive);
int GetExeDir(HINSTANCE hInstance, wchar_t *wszExeDir, int nLen);
int GetFileDir(const wchar_t *wpFile, int nFileLen, wchar_t *wszFileDir, DWORD dwFileDirLen);
const wchar_t* GetFileName(const wchar_t *wpFile, int nFileLen);
int GetBaseName(const wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen);
const wchar_t* GetFileExt(const wchar_t *wpFile, int nFileLen);
int IsFile(const wchar_t *wpFile);
BOOL TrimTrailBackSlash(wchar_t *wszDir);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
wchar_t* GetCommandLineParamsWide();
char* GetCommandLineParamsA();
wchar_t* GetCommandLineParamsW();
int GetCommandLineArg(const wchar_t *wpCmdLine, wchar_t *wszArg, int nArgMax, const wchar_t **wpNextArg);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);

//Global variables
HINSTANCE hInstance;
HICON hDialogIcon;
HWND hMainDlg;
BOOL bShowDlg=TRUE;
LANGID wLangSystem;
HMENU hMenuList;
RECT rcBatchWizardModMinMaxDialog={443, 509, 0, 0};
RECT rcBatchWizardModCurrentDialog={0};
RECT rcTemplatesMinMaxDialog={443, 509, 0, 0};
RECT rcTemplatesCurrentDialog={0};
wchar_t wszBuf[BUFFER_SIZE];
wchar_t wszCmdLine[COMMANDLINE_SIZE];
wchar_t wszExeDir[MAX_PATH];
wchar_t wszInputScriptFile[MAX_PATH];
wchar_t wszOutputDir[MAX_PATH];
wchar_t wszFilterFind[MAX_PATH];
wchar_t wszFilterReplace[MAX_PATH];
wchar_t wszVirtualDubExe[MAX_PATH];
wchar_t *wszTemplate=NULL;
BOOL bFilterMatchCase=FALSE;
BOOL bFilterRegExp=FALSE;
DWORD dwScriptType=STYPE_VIRTUALDUB;
DWORD dwDetectState=DSTATE_NONE;
int nJobCount=0;
INT_PTR nTemplateLen=0;
BOOL bNoParse=FALSE;
BOOL bNoMessageBox=FALSE;

//GCC
#ifdef __GNUC__
int main()
{
  return 0;
}
#endif

//Entry point
void _WinMain()
{
  MSG msg;
  BOOL bMsgStatus;

  WideInitialize();
  InitCommonControls();
  hInstance=GetModuleHandle(NULL);
  wLangSystem=PRIMARYLANGID(GetUserDefaultLangID());
  hDialogIcon=LoadIconA(hInstance, MAKEINTRESOURCEA(IDI_ICON));
  GetExeDir(hInstance, wszExeDir, MAX_PATH);
  GetCommandLineParamsWide();

  if (hMainDlg=CreateDialogWide(hInstance, MAKEINTRESOURCEW(IDD_MAIN), NULL, (DLGPROC)MainDlgProc))
  {
    if (bShowDlg) ShowWindow(hMainDlg, SW_SHOW);
  }

  if (hMainDlg)
  {
    while ((bMsgStatus=GetMessageWide(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (!IsDialogMessageWide(hMainDlg, &msg))
      {
        TranslateMessage(&msg);
        DispatchMessageWide(&msg);
      }
    }
  }

  ExitProcess(0);
}

LRESULT CALLBACK MainDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTemplateGroup;
  static HWND hWndScriptFile;
  static HWND hWndScriptBrowse;
  static HWND hWndScriptParse;
  static HWND hWndTemplate;
  static HWND hWndOutputDirLabel;
  static HWND hWndOutputDir;
  static HWND hWndOutputBrowse;
  static HWND hWndListLabel;
  static HWND hWndList;
  static HWND hWndFilterGroup;
  static HWND hWndFilterFindLabel;
  static HWND hWndFilterFind;
  static HWND hWndFilterReplaceLabel;
  static HWND hWndFilterReplace;
  static HWND hWndFilterMatchCase;
  static HWND hWndFilterRegExp;
  static HWND hWndExecute;
  static HWND hWndScriptSave;
  static HWND hWndClose;
  static int nSelItem=-1;
  static DIALOGRESIZE drs[]={{&hWndTemplateGroup,      DRS_SIZE|DRS_X, 0},
                             {&hWndScriptFile,         DRS_SIZE|DRS_X, 0},
                             {&hWndScriptBrowse,       DRS_MOVE|DRS_X, 0},
                             {&hWndScriptParse,        DRS_MOVE|DRS_X, 0},
                             {&hWndTemplate,           DRS_SIZE|DRS_X, 0},
                             {&hWndListLabel,          DRS_SIZE|DRS_X, 0},
                             {&hWndList,               DRS_SIZE|DRS_X, 0},
                             {&hWndList,               DRS_SIZE|DRS_Y, 0},
                             {&hWndOutputDirLabel,     DRS_SIZE|DRS_X, 0},
                             {&hWndOutputDirLabel,     DRS_MOVE|DRS_Y, 0},
                             {&hWndOutputDir,          DRS_SIZE|DRS_X, 0},
                             {&hWndOutputDir,          DRS_MOVE|DRS_Y, 0},
                             {&hWndOutputBrowse,       DRS_MOVE|DRS_X, 0},
                             {&hWndOutputBrowse,       DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterGroup,        DRS_SIZE|DRS_X, 0},
                             {&hWndFilterGroup,        DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterFindLabel,    DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterFind,         DRS_SIZE|DRS_X, 0},
                             {&hWndFilterFind,         DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterReplaceLabel, DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterReplace,      DRS_SIZE|DRS_X, 0},
                             {&hWndFilterReplace,      DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterMatchCase,    DRS_MOVE|DRS_X, 0},
                             {&hWndFilterMatchCase,    DRS_MOVE|DRS_Y, 0},
                             {&hWndFilterRegExp,       DRS_MOVE|DRS_X, 0},
                             {&hWndFilterRegExp,       DRS_MOVE|DRS_Y, 0},
                             {&hWndExecute,            DRS_MOVE|DRS_Y, 0},
                             {&hWndScriptSave,         DRS_MOVE|DRS_Y, 0},
                             {&hWndClose,              DRS_MOVE|DRS_X, 0},
                             {&hWndClose,              DRS_MOVE|DRS_Y, 0},
                             {0, 0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNW lvcW;

    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hDialogIcon);
    hWndTemplateGroup=GetDlgItem(hDlg, IDC_TEMPLATEGROUP);
    hWndScriptFile=GetDlgItem(hDlg, IDC_SCRIPTFILE);
    hWndScriptBrowse=GetDlgItem(hDlg, IDC_SCRIPTBROWSE);
    hWndScriptParse=GetDlgItem(hDlg, IDC_SCRIPTPARSE);
    hWndTemplate=GetDlgItem(hDlg, IDC_TEMPLATE);
    hWndOutputDirLabel=GetDlgItem(hDlg, IDC_OUTPUTDIR_LABEL);
    hWndOutputDir=GetDlgItem(hDlg, IDC_OUTPUTDIR);
    hWndOutputBrowse=GetDlgItem(hDlg, IDC_OUTPUTBROWSE);
    hWndListLabel=GetDlgItem(hDlg, IDC_LIST_LABEL);
    hWndList=GetDlgItem(hDlg, IDC_LIST);
    hWndFilterGroup=GetDlgItem(hDlg, IDC_FILTERGROUP);
    hWndFilterFindLabel=GetDlgItem(hDlg, IDC_FILTERFIND_LABEL);
    hWndFilterFind=GetDlgItem(hDlg, IDC_FILTERFIND);
    hWndFilterReplaceLabel=GetDlgItem(hDlg, IDC_FILTERREPLACE_LABEL);
    hWndFilterReplace=GetDlgItem(hDlg, IDC_FILTERREPLACE);
    hWndFilterMatchCase=GetDlgItem(hDlg, IDC_FILTERMATCHCASE);
    hWndFilterRegExp=GetDlgItem(hDlg, IDC_FILTERREGEXP);
    hWndExecute=GetDlgItem(hDlg, IDC_EXECUTE);
    hWndScriptSave=GetDlgItem(hDlg, IDC_SCRIPTSAVE);
    hWndClose=GetDlgItem(hDlg, IDC_CLOSE);

    SetWindowTextWide(hDlg, STR_DIALOGTITLE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT, LVS_EX_FULLROWSELECT);
    SetDlgItemTextWide(hDlg, IDC_TEMPLATEGROUP, GetLangStringW(wLangSystem, STRID_CREATETEMPLATE));
    SetDlgItemTextWide(hDlg, IDC_SCRIPTFILE_LABEL, GetLangStringW(wLangSystem, STRID_SCRIPTFILE));
    SetDlgItemTextWide(hDlg, IDC_SCRIPTPARSE, GetLangStringW(wLangSystem, STRID_GENERATE));
    SetDlgItemTextWide(hDlg, IDC_LIST_LABEL, GetLangStringW(wLangSystem, STRID_LISTLABEL));
    SetDlgItemTextWide(hDlg, IDC_OUTPUTDIR_LABEL, GetLangStringW(wLangSystem, STRID_OUTPUTDIR));
    SetDlgItemTextWide(hDlg, IDC_FILTERGROUP, GetLangStringW(wLangSystem, STRID_OUTPUTFILTER));
    SetDlgItemTextWide(hDlg, IDC_FILTERFIND_LABEL, GetLangStringW(wLangSystem, STRID_FINDIT));
    SetDlgItemTextWide(hDlg, IDC_FILTERREPLACE_LABEL, GetLangStringW(wLangSystem, STRID_REPLACEWITH));
    SetDlgItemTextWide(hDlg, IDC_FILTERMATCHCASE, GetLangStringW(wLangSystem, STRID_MATCHCASE));
    SetDlgItemTextWide(hDlg, IDC_FILTERREGEXP, GetLangStringW(wLangSystem, STRID_REGEXP));
    SetDlgItemTextWide(hDlg, IDC_EXECUTE, GetLangStringW(wLangSystem, STRID_EXECUTE));
    SetDlgItemTextWide(hDlg, IDC_SCRIPTSAVE, GetLangStringW(wLangSystem, STRID_SAVE));
    SetDlgItemTextWide(hDlg, IDC_CLOSE, GetLangStringW(wLangSystem, STRID_CLOSE));

    //Columns
    lvcW.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvcW.pszText=GetLangStringW(wLangSystem, STRID_INPUTFILE);
    lvcW.cx=253;
    lvcW.iSubItem=LVSI_INPUTFILE;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVSI_INPUTFILE, (LPARAM)&lvcW);

    lvcW.pszText=GetLangStringW(wLangSystem, STRID_OUTPUTFILENAME);
    lvcW.cx=125;
    lvcW.iSubItem=LVSI_OUTPUTFILENAME;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVSI_OUTPUTFILENAME, (LPARAM)&lvcW);

    //Popup menu
    if (hMenuList=CreatePopupMenu())
    {
      AppendMenuWide(hMenuList, MF_STRING, IDC_ITEMDELETE, GetLangStringW(wLangSystem, STRID_MENU_ITEMDELETE));
      AppendMenuWide(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuWide(hMenuList, MF_STRING, IDC_ITEMMOVEUP, GetLangStringW(wLangSystem, STRID_MENU_ITEMMOVEUP));
      AppendMenuWide(hMenuList, MF_STRING, IDC_ITEMMOVEDOWN, GetLangStringW(wLangSystem, STRID_MENU_ITEMMOVEDOWN));
      AppendMenuWide(hMenuList, MF_SEPARATOR, (UINT)-1, NULL);
      AppendMenuWide(hMenuList, MF_STRING, IDC_ITEMCLEAR, GetLangStringW(wLangSystem, STRID_MENU_ITEMCLEAR));
    }
    DragAcceptFiles(hDlg, TRUE);

    wszInputScriptFile[0]=L'\0';
    wszOutputDir[0]=L'\0';
    wszFilterFind[0]=L'\0';
    wszFilterReplace[0]=L'\0';
    wszVirtualDubExe[0]=L'\0';

    if (bFilterMatchCase)
      SendMessage(hWndFilterMatchCase, BM_SETCHECK, BST_CHECKED, 0);
    if (bFilterRegExp)
      SendMessage(hWndFilterRegExp, BM_SETCHECK, BST_CHECKED, 0);

    //Add files from command line
    {
      wchar_t wpOutputScriptFile[MAX_PATH];
      wchar_t wszArg[MAX_PATH];
      wchar_t *pArguments=wszCmdLine;
      int nAdded=0;
      BOOL bParseScript=FALSE;
      BOOL bSaveScript=FALSE;
      BOOL bExecute=FALSE;
      BOOL bQuit=FALSE;

      while (GetCommandLineArg(pArguments, wszArg, MAX_PATH, &pArguments))
      {
        if (wszArg[0] == L'/')
        {
          if (!xstrcmpinW(L"/?", wszArg, (UINT_PTR)-1))
          {
            MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_HELP), STR_DIALOGTITLE, MB_OK|MB_ICONINFORMATION);
            bShowDlg=FALSE;
            PostMessage(hDlg, WM_CLOSE, 0, 0);
            return 1;
          }
          else if (!xstrcmpinW(L"/S=", wszArg, (UINT_PTR)-1))
          {
            GetCommandLineArg(wszArg + xstrlenW(L"/S="), wszInputScriptFile, MAX_PATH, NULL);
            SetWindowTextWide(hWndScriptFile, wszInputScriptFile);
          }
          else if (!xstrcmpinW(L"/P", wszArg, (UINT_PTR)-1))
          {
            bParseScript=TRUE;
          }
          else if (!xstrcmpinW(L"/O=", wszArg, (UINT_PTR)-1))
          {
            GetCommandLineArg(wszArg + xstrlenW(L"/O="), wszOutputDir, MAX_PATH, NULL);
            SetWindowTextWide(hWndOutputDir, wszOutputDir);
          }
          else if (!xstrcmpinW(L"/FF=", wszArg, (UINT_PTR)-1))
          {
            GetCommandLineArg(wszArg + xstrlenW(L"/FF="), wszFilterFind, MAX_PATH, NULL);
            SetWindowTextWide(hWndFilterFind, wszFilterFind);
          }
          else if (!xstrcmpinW(L"/FR=", wszArg, (UINT_PTR)-1))
          {
            GetCommandLineArg(wszArg + xstrlenW(L"/FR="), wszFilterReplace, MAX_PATH, NULL);
            SetWindowTextWide(hWndFilterReplace, wszFilterReplace);
          }
          else if (!xstrcmpinW(L"/FC", wszArg, (UINT_PTR)-1))
          {
            bFilterMatchCase=TRUE;
            SendMessage(hWndFilterMatchCase, BM_SETCHECK, BST_CHECKED, 0);
          }
          else if (!xstrcmpinW(L"/FE", wszArg, (UINT_PTR)-1))
          {
            bFilterRegExp=TRUE;
            SendMessage(hWndFilterRegExp, BM_SETCHECK, BST_CHECKED, 0);
          }
          else if (!xstrcmpinW(L"/SAVE=", wszArg, (UINT_PTR)-1))
          {
            GetCommandLineArg(wszArg + xstrlenW(L"/SAVE="), wpOutputScriptFile, MAX_PATH, NULL);
            bSaveScript=TRUE;
          }
          else if (!xstrcmpinW(L"/EXEC", wszArg, (UINT_PTR)-1))
          {
            bExecute=TRUE;
          }
          else if (!xstrcmpinW(L"/NOMSG", wszArg, (UINT_PTR)-1))
          {
            bNoMessageBox=TRUE;
          }
          else if (!xstrcmpinW(L"/QUIT", wszArg, (UINT_PTR)-1))
          {
            bQuit=TRUE;
          }
        }
        else if (IsFile(wszArg) == ERROR_SUCCESS)
        {
          ListSetItem(hWndList, -1, wszArg, GetFileName(wszArg, -1), TRUE);
          ++nAdded;
        }
      }
      if (nAdded) SendMessage(hWndList, LVM_SCROLL, 0, 0x0FFFFFFF);
      if (bParseScript) SendMessage(hDlg, WM_COMMAND, IDC_SCRIPTPARSE, 0);
      if (bSaveScript) SaveScript(hDlg, hWndList, wpOutputScriptFile);
      if (bExecute) PostMessage(hDlg, WM_COMMAND, IDC_EXECUTE, 0);
      if (bQuit) PostMessage(hDlg, WM_CLOSE, 0, 0);
    }
    SendMessage(hDlg, WM_COMMAND, IDC_UPDATEACCESS, 0);
    SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
  }
  else if (uMsg == WM_DROPFILES)
  {
    wchar_t wszFile[MAX_PATH];
    int nDropped;
    int nAdded=0;
    int i;

    nDropped=DragQueryFileWide((HDROP)wParam, 0xFFFFFFFF, NULL, 0);

    for (i=0; i < nDropped; ++i)
    {
      DragQueryFileWide((HDROP)wParam, i, wszFile, MAX_PATH);
      if (IsFile(wszFile) == ERROR_SUCCESS)
      {
        ListSetItem(hWndList, -1, wszFile, GetFileName(wszFile, -1), TRUE);
        ++nAdded;
      }
    }
    DragFinish((HDROP)wParam);

    if (nAdded) SendMessage(hWndList, LVM_SCROLL, 0, 0x0FFFFFFF);
    SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
    return TRUE;
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hWndList)
    {
      LVHITTESTINFO lvhti;
      LVITEMW lvi;
      POINT ptScreen={0};
      RECT rcItem;
      int nCountItem;

      if (lParam == -1)
      {
        if ((lvhti.iItem=nSelItem) != -1)
        {
          rcItem.left=LVIR_LABEL;
          SendMessage(hWndList, LVM_GETITEMRECT, (WPARAM)nSelItem, (LPARAM)&rcItem);
          ptScreen.x=rcItem.left;
          ptScreen.y=rcItem.bottom;
        }
        ClientToScreen(hWndList, &ptScreen);
      }
      else
      {
        GetCursorPos(&ptScreen);
        lvhti.pt=ptScreen;
        ScreenToClient(hWndList, &lvhti.pt);
        SendMessage(hWndList, LVM_SUBITEMHITTEST, 0, (LPARAM)&lvhti);

        lvi.stateMask=LVIS_SELECTED;
        lvi.state=LVIS_SELECTED;
        SendMessage(hWndList, LVM_SETITEMSTATE, (WPARAM)lvhti.iItem, (LPARAM)&lvi);
      }

      nCountItem=(int)SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0);
      EnableMenuItem(hMenuList, IDC_ITEMDELETE, nSelItem >= 0?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hMenuList, IDC_ITEMMOVEUP, (nSelItem >= 0 && nSelItem > 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hMenuList, IDC_ITEMMOVEDOWN, (nSelItem >= 0 && nSelItem < nCountItem - 1)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hMenuList, IDC_ITEMCLEAR, nCountItem > 0?MF_ENABLED:MF_GRAYED);
      TrackPopupMenu(hMenuList, TPM_LEFTBUTTON|TPM_RIGHTBUTTON, ptScreen.x, ptScreen.y, 0, hDlg, NULL);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_UPDATEACCESS)
    {
      BOOL bOutputEnable=(dwDetectState & DSTATE_SINGLEFILESINGLEPASS) && !(dwDetectState & DSTATE_NOOUTPUTFILE);

      if (dwScriptType & STYPE_AVS)
        SetWindowTextWide(hWndExecute, GetLangStringW(wLangSystem, STRID_CREATE));
      else
        SetWindowTextWide(hWndExecute, GetLangStringW(wLangSystem, STRID_EXECUTE));

      if (IsFile(wszInputScriptFile) == ERROR_SUCCESS)
        EnableWindow(hWndScriptParse, TRUE);
      else
        EnableWindow(hWndScriptParse, FALSE);

      if (!wszTemplate)
      {
        SetWindowTextWide(hWndTemplate, GetLangStringW(wLangSystem, STRID_EMPTY));
        bOutputEnable=FALSE;
      }
      else
      {
        if (dwScriptType & STYPE_AVS)
          bOutputEnable=TRUE;
        SetWindowTextWide(hWndTemplate, GetLangStringW(wLangSystem, STRID_TEMPLATE));
      }

      EnableWindow(hWndExecute, (BOOL)(INT_PTR)wszTemplate);
      EnableWindow(hWndScriptSave, (BOOL)((dwScriptType & STYPE_VIRTUALDUB) && wszTemplate));

      EnableWindow(hWndOutputDirLabel, bOutputEnable);
      EnableWindow(hWndOutputDir, bOutputEnable);
      EnableWindow(hWndOutputBrowse, bOutputEnable);
      EnableWindow(hWndFilterGroup, bOutputEnable);
      EnableWindow(hWndFilterFindLabel, bOutputEnable);
      EnableWindow(hWndFilterFind, bOutputEnable);
      EnableWindow(hWndFilterReplaceLabel, bOutputEnable);
      EnableWindow(hWndFilterReplace, bOutputEnable);
      EnableWindow(hWndFilterMatchCase, bOutputEnable);
      EnableWindow(hWndFilterRegExp, bOutputEnable);
    }
    else if (LOWORD(wParam) == IDC_UPDATEFILTER)
    {
      LVITEMW lvi;
      wchar_t wszInputFile[MAX_PATH];
      wchar_t *wszFilterFileName;
      const wchar_t *wpFileName;
      const wchar_t *wpExt;
      INT_PTR nFilterFileNameLen;
      int nIndex=0;

      for (;;)
      {
        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszInputFile;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nIndex;
        lvi.iSubItem=LVSI_INPUTFILE;
        if (!ListView_GetItemWide(hWndList, &lvi))
          break;

        wpFileName=GetFileName(wszInputFile, -1);
        nFilterFileNameLen=xstrlenW(wpFileName);
        if (wszFilterFileName=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nFilterFileNameLen + 1) * sizeof(wchar_t)))
          xstrcpynW(wszFilterFileName, wpFileName, nFilterFileNameLen + 1);

        if (dwScriptType & STYPE_AVS)
        {
          if (wpExt=GetFileExt(wszFilterFileName, -1))
            TextReplaceRE(&wszFilterFileName, &nFilterFileNameLen, L"[^\\.]*$", -1, L"avs", -1, TRUE);
          else
            TextReplaceRE(&wszFilterFileName, &nFilterFileNameLen, L"\\Z", -1, L"\\.avs", -1, TRUE);
        }

        if (bFilterRegExp)
        {
          TextReplaceRE(&wszFilterFileName, &nFilterFileNameLen, wszFilterFind, -1, wszFilterReplace, -1, bFilterMatchCase);
        }
        else
        {
          //Include NULL symbol for TextReplace
          ++nFilterFileNameLen;

          TextReplace(&wszFilterFileName, &nFilterFileNameLen, wszFilterFind, -1, wszFilterReplace, -1, bFilterMatchCase);

          //Exclude NULL symbol
          --nFilterFileNameLen;
        }

        ListSetItem(hWndList, nIndex, NULL, wszFilterFileName, FALSE);

        GlobalFree((HGLOBAL)wszFilterFileName);
        ++nIndex;
      }
    }
    else if (LOWORD(wParam) == IDC_SCRIPTFILE)
    {
      GetWindowTextWide(hWndScriptFile, wszInputScriptFile, MAX_PATH);
      SendMessage(hDlg, WM_COMMAND, IDC_UPDATEACCESS, 0);
    }
    else if (LOWORD(wParam) == IDC_SCRIPTBROWSE)
    {
      OPENFILENAMEW efn;

      xmemset(&efn, 0, sizeof(OPENFILENAMEW));
      efn.lStructSize  =sizeof(OPENFILENAMEW);
      efn.hwndOwner    =hDlg;
      efn.lpstrFile    =wszInputScriptFile;
      efn.nMaxFile     =MAX_PATH;
      efn.lpstrFilter  =L"*.jobs;*.vcf;*.avs;*.avst\0*.jobs;*.vcf;*.avs;*.avst\0*.*\0*.*\0\0";
      efn.nFilterIndex =1;
      efn.Flags        =OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;

      if (GetOpenFileNameWide(&efn))
      {
        SetWindowTextWide(hWndScriptFile, wszInputScriptFile);
        SendMessage(hDlg, WM_COMMAND, IDC_UPDATEACCESS, 0);
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SCRIPTPARSE)
    {
      if (wszTemplate)
      {
        GlobalFree((HGLOBAL)wszTemplate);
        wszTemplate=NULL;
      }
      nJobCount=0;
      dwDetectState=DSTATE_NONE;
      dwScriptType=STYPE_VIRTUALDUB;

      if (IsFile(wszInputScriptFile) == ERROR_SUCCESS)
      {
        const wchar_t *wpExt;
        wchar_t *wpContent=NULL;
        DWORD dwContentLen;

        if (wpExt=GetFileExt(wszInputScriptFile, -1))
        {
          if (!xstrcmpiW(wpExt, L"avs"))
            dwScriptType=STYPE_AVS;
          else if (!xstrcmpiW(wpExt, L"avst"))
            dwScriptType=STYPE_AVS|STYPE_AVST;
        }
        if (ComboBox_FindStringExactWide(hWndScriptFile, 0, wszInputScriptFile) == CB_ERR)
          ComboBox_InsertStringWide(hWndScriptFile, 0, wszInputScriptFile);

        if (dwContentLen=ReadAnsiFileAsUnicode(wszInputScriptFile, &wpContent))
        {
          ParseScript(hDlg, wpContent, dwContentLen, &wszTemplate, &nTemplateLen, FALSE);
          GlobalFree((HGLOBAL)wpContent);
        }
      }
      SendMessage(hDlg, WM_COMMAND, IDC_UPDATEACCESS, 0);
      SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
    }
    else if (LOWORD(wParam) == IDC_TEMPLATE)
    {
      if (DialogBoxWide(hInstance, MAKEINTRESOURCEW(IDD_TEMPLATE), hDlg, (DLGPROC)TemplateDlgProc))
      {
        wchar_t *wpContent=wszTemplate;
        INT_PTR nContentLen=nTemplateLen;

        nJobCount=0;
        dwDetectState=DSTATE_NONE;

        if (wpContent)
        {
          ParseScript(hDlg, wpContent, nContentLen, &wszTemplate, &nTemplateLen, bNoParse);
          GlobalFree((HGLOBAL)wpContent);
        }
        SendMessage(hDlg, WM_COMMAND, IDC_UPDATEACCESS, 0);
        SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OUTPUTBROWSE)
    {
      BROWSEINFOW biW;
      LPITEMIDLIST pIdList;
      LPMALLOC pMalloc;

      GetWindowTextWide(hWndOutputDir, wszOutputDir, MAX_PATH);
      biW.hwndOwner=hDlg;
      biW.pidlRoot=NULL;
      biW.pszDisplayName=wszOutputDir;
      biW.lpszTitle=NULL;
      biW.ulFlags=BIF_RETURNONLYFSDIRS|BIF_NEWDIALOGSTYLE|BIF_EDITBOX;
      biW.lpfn=BrowseCallbackProc;
      biW.lParam=(LPARAM)wszOutputDir;
      biW.iImage=0;

      if (pIdList=SHBrowseForFolderWide(&biW))
      {
        SHGetPathFromIDListWide(pIdList, wszOutputDir);

        if (SHGetMalloc(&pMalloc))
        {
          pMalloc->lpVtbl->Free(pMalloc, pIdList);
          pMalloc->lpVtbl->Release(pMalloc);
        }
        SetWindowTextWide(hWndOutputDir, wszOutputDir);
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_FILTERFIND ||
             LOWORD(wParam) == IDC_FILTERREPLACE)
    {
      if (HIWORD(wParam) == EN_CHANGE)
      {
        GetWindowTextWide(hWndFilterFind, wszFilterFind, MAX_PATH);
        GetWindowTextWide(hWndFilterReplace, wszFilterReplace, MAX_PATH);
        SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
      }
    }
    else if (LOWORD(wParam) == IDC_FILTERMATCHCASE)
    {
      bFilterMatchCase=SendMessage(hWndFilterMatchCase, BM_GETCHECK, 0, 0);
      SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
    }
    else if (LOWORD(wParam) == IDC_FILTERREGEXP)
    {
      bFilterRegExp=SendMessage(hWndFilterRegExp, BM_GETCHECK, 0, 0);
      SendMessage(hDlg, WM_COMMAND, IDC_UPDATEFILTER, 0);
    }
    else if (LOWORD(wParam) == IDC_EXECUTE)
    {
      wchar_t wpOutputScriptFile[MAX_PATH];

      GetWindowTextWide(hWndOutputDir, wszOutputDir, MAX_PATH);
      if (!(dwDetectState & DSTATE_NOOUTPUTFILE))
      {
        if (!*wszOutputDir && (dwScriptType & STYPE_AVS))
        {
          xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_OUTPUTDIREMPTY), wszOutputDir);
          if (!bNoMessageBox && MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_YESNO|MB_ICONQUESTION) == IDNO)
            return FALSE;
        }
        else if (IsFile(wszOutputDir) != ERROR_DIRECTORY)
        {
          xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_OUTPUTDIRNOTFOUND), wszOutputDir);
          if (!bNoMessageBox) MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_OK|MB_ICONERROR);
          return FALSE;
        }
      }
      TrimTrailBackSlash(wszOutputDir);

      if (dwScriptType & STYPE_VIRTUALDUB)
      {
        if (!*wszVirtualDubExe || IsFile(wszVirtualDubExe) != ERROR_SUCCESS)
        {
          xprintfW(wszVirtualDubExe, L"%s\\VirtualDubMod.exe", wszExeDir);
          if (IsFile(wszVirtualDubExe) != ERROR_SUCCESS)
          {
            xprintfW(wszVirtualDubExe, L"%s\\VirtualDub.exe", wszExeDir);
            if (IsFile(wszVirtualDubExe) != ERROR_SUCCESS)
            {
              OPENFILENAMEW efn;

              xmemset(&efn, 0, sizeof(OPENFILENAMEW));
              efn.lStructSize  =sizeof(OPENFILENAMEW);
              efn.hwndOwner    =hDlg;
              efn.lpstrFile    =wszVirtualDubExe;
              efn.nMaxFile     =MAX_PATH;
              efn.lpstrFilter  =L"VirtualDubMod.exe;VirtualDub.exe\0VirtualDubMod.exe;VirtualDub.exe\0*.exe\0*.exe\0\0";
              efn.nFilterIndex =1;
              efn.Flags        =OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;

              if (!GetOpenFileNameWide(&efn))
              {
                if (!bNoMessageBox) MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_VIRTUALDUBNOTFOUND), STR_DIALOGTITLE, MB_OK|MB_ICONERROR);
                return FALSE;
              }
            }
          }
        }

        if (GetTempPathWide(MAX_PATH, wszBuf))
        {
          if (GetTempFileNameWide(wszBuf, L"bwm", 0, wpOutputScriptFile))
          {
            if (SaveScript(hDlg, hWndList, wpOutputScriptFile) > 0)
            {
              STARTUPINFOW si;
              PROCESS_INFORMATION pi;

              xprintfW(wszBuf, L"\"%s\" /s\"%s\"", wszVirtualDubExe, wpOutputScriptFile);
              xmemset(&si, 0, sizeof(STARTUPINFOW));
              si.cb=sizeof(STARTUPINFOW);
              if (CreateProcessWide(NULL, wszBuf, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
              {
                CloseHandle(pi.hProcess);
                CloseHandle(pi.hThread);
              }
            }
          }
        }
      }
      else if (dwScriptType & STYPE_AVS)
      {
        int nFilesCount;

        if ((nFilesCount=SaveScript(hDlg, hWndList, NULL)) >= 0)
        {
          xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_DONEFILES), nFilesCount);
          if (!bNoMessageBox) MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_OK|MB_ICONINFORMATION);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_SCRIPTSAVE)
    {
      OPENFILENAMEW ofn;
      wchar_t wpOutputScriptFile[MAX_PATH];

      GetWindowTextWide(hWndOutputDir, wszOutputDir, MAX_PATH);
      if (!(dwDetectState & DSTATE_NOOUTPUTFILE))
      {
        if (IsFile(wszOutputDir) != ERROR_DIRECTORY)
        {
          xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_OUTPUTDIRNOTFOUNDCONTINUE), wszOutputDir);
          if (!bNoMessageBox && MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON2|MB_ICONEXCLAMATION) == IDNO)
            return FALSE;
        }
      }
      TrimTrailBackSlash(wszOutputDir);

      //Choose file
      xstrcpynW(wpOutputScriptFile, wszInputScriptFile, MAX_PATH);
      xmemset(&ofn, 0, sizeof(OPENFILENAMEW));
      ofn.lStructSize  =sizeof(OPENFILENAMEW);
      ofn.hwndOwner    =hDlg;
      ofn.hInstance    =hInstance;
      ofn.lpstrFile    =wpOutputScriptFile;
      ofn.lpstrFilter  =L"*.jobs\0*.jobs\0*.*\0*.*\0\0";
      ofn.nFilterIndex =1;
      ofn.lpstrDefExt  =L"jobs";
      ofn.nMaxFile     =MAX_PATH;
      ofn.Flags        =OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_EXPLORER|OFN_ENABLESIZING|OFN_OVERWRITEPROMPT;

      if (GetSaveFileNameWide(&ofn))
        SaveScript(hDlg, hWndList, wpOutputScriptFile);
    }
    else if (LOWORD(wParam) == IDC_ITEMMOVEUP ||
             LOWORD(wParam) == IDC_ITEMMOVEDOWN)
    {
      if (nSelItem != -1)
      {
        LVITEMW lvi;
        wchar_t wszInputFile[MAX_PATH];
        wchar_t wszOutputFileName[MAX_PATH];
        int nOldIndex=nSelItem;
        int nNewIndex;

        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszInputFile;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nOldIndex;
        lvi.iSubItem=LVSI_INPUTFILE;
        ListView_GetItemWide(hWndList, &lvi);

        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszOutputFileName;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nOldIndex;
        lvi.iSubItem=LVSI_OUTPUTFILENAME;
        ListView_GetItemWide(hWndList, &lvi);

        //Delete
        SendMessage(hWndList, LVM_DELETEITEM, nOldIndex, 0);

        //Add
        if (LOWORD(wParam) == IDC_ITEMMOVEUP)
          nNewIndex=max(nOldIndex - 1, 0);
        else
          nNewIndex=nOldIndex + 1;
        nNewIndex=ListSetItem(hWndList, nNewIndex, wszInputFile, wszOutputFileName, TRUE);

        //Select
        SetFocus(hWndList);
        ListSelItem(hWndList, nNewIndex);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEMDELETE)
    {
      if (nSelItem != -1)
      {
        int nIndex=nSelItem;

        //Delete
        SendMessage(hWndList, LVM_DELETEITEM, nIndex, 0);
        nIndex=min(nIndex, (int)SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0) - 1);

        //Select
        SetFocus(hWndList);
        ListSelItem(hWndList, nIndex);
      }
    }
    else if (LOWORD(wParam) == IDC_ITEMCLEAR)
    {
      SendMessage(hWndList, LVM_DELETEALLITEMS, 0, 0);
    }
    else if (LOWORD(wParam) == IDC_CLOSE || LOWORD(wParam) == IDCANCEL)
    {
      PostMessage(hDlg, WM_CLOSE, 0, 0);
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_LIST)
    {
      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          nSelItem=((NMLISTVIEW *)lParam)->iItem;
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          nSelItem=-1;
        }
      }
      else if (((NMHDR *)lParam)->code == (UINT)LVN_KEYDOWN)
      {
        NMLVKEYDOWN *pnkd=(NMLVKEYDOWN *)lParam;
        BOOL bAlt=FALSE;
        BOOL bShift=FALSE;
        BOOL bControl=FALSE;

        if (GetKeyState(VK_MENU) < 0)
          bAlt=TRUE;
        if (GetKeyState(VK_SHIFT) < 0)
          bShift=TRUE;
        if (GetKeyState(VK_CONTROL) < 0)
          bControl=TRUE;

        if (pnkd->wVKey == VK_DELETE)
        {
          if (!bAlt && !bShift && !bControl)
          {
            PostMessage(hDlg, WM_COMMAND, IDC_ITEMDELETE, 0);
            return TRUE;
          }
        }
        else if (pnkd->wVKey == VK_UP)
        {
          if (bAlt && !bShift && !bControl)
          {
            PostMessage(hDlg, WM_COMMAND, IDC_ITEMMOVEUP, 0);
            return TRUE;
          }
        }
        else if (pnkd->wVKey == VK_DOWN)
        {
          if (bAlt && !bShift && !bControl)
          {
            PostMessage(hDlg, WM_COMMAND, IDC_ITEMMOVEDOWN, 0);
            return TRUE;
          }
        }
      }
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    if (wszTemplate)
    {
      GlobalFree((HGLOBAL)wszTemplate);
      wszTemplate=NULL;
    }
    DestroyMenu(hMenuList);
    DestroyWindow(hMainDlg);
    hMainDlg=NULL;
  }
  else if (uMsg == WM_DESTROY)
  {
    PostQuitMessage(0);
  }
  DialogResizeMessages(&drs[0], &rcBatchWizardModMinMaxDialog, &rcBatchWizardModCurrentDialog, DRM_PAINTSIZEGRIP, hDlg, uMsg, wParam, lParam);

  return 0;
}

BOOL CALLBACK TemplateDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTypeGroup;
  static HWND hWndTypeVirtualDub;
  static HWND hWndTypeAVS;
  static HWND hWndEdit;
  static HWND hWndNoParse;
  static HWND hWndOK;
  static HWND hWndCancel;
  static DIALOGRESIZE drs[]={{&hWndTypeGroup, DRS_SIZE|DRS_X, 0},
                             {&hWndTypeAVS,   DRS_MOVE|DRS_X, 0},
                             {&hWndEdit,      DRS_SIZE|DRS_X, 0},
                             {&hWndEdit,      DRS_SIZE|DRS_Y, 0},
                             {&hWndNoParse,   DRS_MOVE|DRS_Y, 0},
                             {&hWndOK,        DRS_MOVE|DRS_X, 0},
                             {&hWndOK,        DRS_MOVE|DRS_Y, 0},
                             {&hWndCancel,    DRS_MOVE|DRS_X, 0},
                             {&hWndCancel,    DRS_MOVE|DRS_Y, 0},
                             {0, 0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hDialogIcon);
    hWndTypeGroup=GetDlgItem(hDlg, IDC_TEMPLATE_TYPEGROUP);
    hWndTypeVirtualDub=GetDlgItem(hDlg, IDC_TEMPLATE_TYPEVIRTUALDUB);
    hWndTypeAVS=GetDlgItem(hDlg, IDC_TEMPLATE_TYPEAVS);
    hWndEdit=GetDlgItem(hDlg, IDC_TEMPLATE_EDIT);
    hWndNoParse=GetDlgItem(hDlg, IDC_TEMPLATE_NOPARSE);
    hWndOK=GetDlgItem(hDlg, IDOK);
    hWndCancel=GetDlgItem(hDlg, IDCANCEL);

    SetWindowTextWide(hDlg, GetLangStringW(wLangSystem, STRID_TEMPLATE));
    SetDlgItemTextWide(hDlg, IDC_TEMPLATE_TYPEGROUP, GetLangStringW(wLangSystem, STRID_SCRIPTTYPE));
    SetDlgItemTextWide(hDlg, IDC_TEMPLATE_TYPEVIRTUALDUB, GetLangStringW(wLangSystem, STRID_SCRIPTVIRTUALDUB));
    SetDlgItemTextWide(hDlg, IDC_TEMPLATE_TYPEAVS, GetLangStringW(wLangSystem, STRID_SCRIPTAVS));
    SetDlgItemTextWide(hDlg, IDC_TEMPLATE_NOPARSE, GetLangStringW(wLangSystem, STRID_NOPARSE));
    SetDlgItemTextWide(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextWide(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    //VitualDub script type is the default one
    if (dwScriptType & STYPE_AVS)
      SendMessage(hWndTypeAVS, BM_SETCHECK, BST_CHECKED, 0);
    else
      SendMessage(hWndTypeVirtualDub, BM_SETCHECK, BST_CHECKED, 0);

    if (bNoParse)
      SendMessage(hWndNoParse, BM_SETCHECK, BST_CHECKED, 0);
    if (wszTemplate)
      SetWindowTextWide(hWndEdit, wszTemplate);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      nTemplateLen=GetWindowTextLengthWide(hWndEdit);

      if (wszTemplate)
      {
        GlobalFree((HGLOBAL)wszTemplate);
        wszTemplate=NULL;
      }
      if (nTemplateLen)
      {
        if (wszTemplate=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTemplateLen + 1) * sizeof(wchar_t)))
          GetWindowTextWide(hWndEdit, wszTemplate, nTemplateLen + 1);
      }

      if (SendMessage(hWndTypeAVS, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwScriptType=STYPE_AVS;
      else
        dwScriptType=STYPE_VIRTUALDUB;

      bNoParse=(BOOL)SendMessage(hWndNoParse, BM_GETCHECK, 0, 0);

      EndDialog(hDlg, 1);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  DialogResizeMessages(&drs[0], &rcTemplatesMinMaxDialog, &rcTemplatesCurrentDialog, DRM_PAINTSIZEGRIP, hDlg, uMsg, wParam, lParam);

  return FALSE;
}

int CALLBACK BrowseCallbackProc(HWND hWnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
  if (uMsg == BFFM_INITIALIZED)
  {
    if (WideGlobal_bOldWindows)
    {
      char szData[MAX_PATH];

      WideCharToMultiByte(CP_ACP, 0, (wchar_t *)lpData, -1, szData, MAX_PATH, NULL, NULL);
      SendMessage(hWnd, BFFM_SETSELECTIONA, TRUE, (LPARAM)szData);
    }
    else SendMessage(hWnd, BFFM_SETSELECTIONW, TRUE, lpData);
  }
  return 0;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcMinMax, RECT *rcCurrent, DWORD dwFlags, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;
    RECT rcControl;
    DWORD dwFlags=SWP_NOMOVE;
    int i;

    rcTemplate=*rcCurrent;
    GetWindowPos(hDlg, NULL, rcCurrent);

    for (i=0; drs[i].lpWnd; ++i)
    {
      if (*drs[i].lpWnd)
      {
        GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
        if (drs[i].dwType & DRS_SIZE)
        {
          if (drs[i].dwType & DRS_X)
            drs[i].nOffset=rcCurrent->right - (rcControl.left + rcControl.right);
          else if (drs[i].dwType & DRS_Y)
            drs[i].nOffset=rcCurrent->bottom - (rcControl.top + rcControl.bottom);
        }
        else if (drs[i].dwType & DRS_MOVE)
        {
          if (drs[i].dwType & DRS_X)
            drs[i].nOffset=rcCurrent->right - rcControl.left;
          else if (drs[i].dwType & DRS_Y)
            drs[i].nOffset=rcCurrent->bottom - rcControl.top;
        }
      }
    }

    if (rcTemplate.right && rcTemplate.bottom)
    {
      if (GetWindowLongPtrWide(hDlg, GWL_STYLE) & DS_CENTER)
      {
        rcTemplate.left=rcCurrent->left + (rcCurrent->right - rcTemplate.right) / 2;
        rcTemplate.top=rcCurrent->top + (rcCurrent->bottom - rcTemplate.bottom) / 2;
        dwFlags&=~SWP_NOMOVE;
      }
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, dwFlags|SWP_NOZORDER|SWP_NOACTIVATE);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    if (rcMinMax)
    {
      MINMAXINFO *mmi=(MINMAXINFO *)lParam;

      if (rcMinMax->left)
        mmi->ptMinTrackSize.x=rcMinMax->left;
      if (rcMinMax->top)
        mmi->ptMinTrackSize.y=rcMinMax->top;
      if (rcMinMax->right)
        mmi->ptMaxTrackSize.x=rcMinMax->right;
      if (rcMinMax->bottom)
        mmi->ptMaxTrackSize.y=rcMinMax->bottom;
    }
  }
  else if (uMsg == WM_MOVE)
  {
    if (!(GetWindowLongPtrWide(hDlg, GWL_STYLE) & DS_CENTER))
    {
      RECT rcTemplate;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      rcCurrent->left=rcTemplate.left;
      rcCurrent->top=rcTemplate.top;
      return TRUE;
    }
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcControl;
      DWORD dwFlags;
      int i;

      GetWindowPos(hDlg, NULL, rcCurrent);

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          else
            continue;

          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, (drs[i].dwType & DRS_X)?(rcCurrent->right - drs[i].nOffset):rcControl.left,
                                         (drs[i].dwType & DRS_Y)?(rcCurrent->bottom - drs[i].nOffset):rcControl.top,
                                         (drs[i].dwType & DRS_X)?(rcCurrent->right - rcControl.left - drs[i].nOffset):rcControl.right,
                                         (drs[i].dwType & DRS_Y)?(rcCurrent->bottom - rcControl.top - drs[i].nOffset):rcControl.bottom,
                                          dwFlags|SWP_NOZORDER|SWP_NOACTIVATE);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    if (dwFlags & DRM_PAINTSIZEGRIP)
    {
      PAINTSTRUCT ps;
      RECT rcGrip;
      HDC hDC;

      if (hDC=BeginPaint(hDlg, &ps))
      {
        GetClientRect(hDlg, &rcGrip);
        rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
        rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
        DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
        EndPaint(hDlg, &ps);
      }
    }
  }
  return FALSE;
}

BOOL ListSelItem(HWND hWnd, int nIndex)
{
  LVITEMA lvi;

  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  return (BOOL)SendMessage(hWnd, LVM_SETITEMSTATE, nIndex, (LPARAM)&lvi);
}

int ListSetItem(HWND hWnd, int nIndex, const wchar_t *wpFile, const wchar_t *wpFileName, BOOL bNew)
{
  LVITEMW lvi;

  if (wpFile)
  {
    lvi.mask=LVIF_TEXT;
    lvi.pszText=(wchar_t *)wpFile;
    lvi.iItem=(nIndex > -1)?nIndex:0x7FFFFFFF;
    lvi.iSubItem=LVSI_INPUTFILE;
    if (bNew)
      nIndex=ListView_InsertItemWide(hWnd, &lvi);
    else
      ListView_SetItemWide(hWnd, &lvi);
  }
  if (wpFileName)
  {
    lvi.mask=LVIF_TEXT;
    lvi.pszText=(wchar_t *)wpFileName;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVSI_OUTPUTFILENAME;
    ListView_SetItemWide(hWnd, &lvi);
  }
  return nIndex;
}

DWORD ReadAnsiFileAsUnicode(const wchar_t *wpFile, wchar_t **wpContent)
{
  unsigned char *szText;
  wchar_t *wszText=NULL;
  HANDLE hFile;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen=0;

  if ((hFile=CreateFileWide(wpFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
  {
    if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
    {
      if (szText=(unsigned char *)GlobalAlloc(GPTR, dwFileSize))
      {
        if (ReadFile(hFile, szText, dwFileSize, &dwBytesRead, NULL))
        {
          if (wszText=(wchar_t *)GlobalAlloc(GPTR, dwBytesRead * sizeof(wchar_t) + sizeof(wchar_t)))
          {
            dwUnicodeLen=MultiByteToWideChar(CP_ACP, 0, szText, dwBytesRead, wszText, dwBytesRead);
            wszText[dwUnicodeLen]=L'\0';
          }
        }
        GlobalFree((HGLOBAL)szText);
      }
    }
    CloseHandle(hFile);
  }
  if (wpContent) *wpContent=wszText;
  return dwUnicodeLen;
}

void ParseScript(HWND hDlg, wchar_t *wpInput, INT_PTR nInputLen, wchar_t **wppOutput, INT_PTR *lpnOutputLen, BOOL bSkipParse)
{
  PATEXEC pe;
  REGROUP *lpREGroupItem;
  wchar_t wszMethod[MAX_PATH];
  wchar_t wszInputFile[MAX_PATH];
  wchar_t wszFileNoSlash[MAX_PATH];
  wchar_t wszFileNoExt[MAX_PATH];
  wchar_t wszExt[32];
  wchar_t *wszOutput=NULL;
  const wchar_t *wpExt;
  const wchar_t *wpScriptStart=wpInput;
  INT_PTR nOutputLen=0;
  INT_PTR nTextLen;
  int nReplaceCount;
  BOOL bVideoFile;
  BOOL bCheckWAV;
  BOOL bOutputDirEmpty=!*wszOutputDir;

  wszInputFile[0]=L'\0';

  if ((dwScriptType & STYPE_AVST) || (dwScriptType & STYPE_AVS))
  {
    nOutputLen=nInputLen;
    if (wszOutput=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nOutputLen + 1) * sizeof(wchar_t)))
      xstrcpynW(wszOutput, wpInput, nOutputLen + 1);

    if ((dwScriptType & STYPE_AVST) || ((dwScriptType & STYPE_AVS) && xstrstrW(wpInput, nInputLen, L"\"%f\"", -1, FALSE, NULL, NULL)))
    {
      //Remove brackets []
      TextReplaceRE(&wszOutput, &nOutputLen, L"(?:\\[([^\\n\"]*))(\"%f\")(?:([^\\n\"]*)\\])", -1, L"\\1\\2\\3", -1, FALSE);

      //Force \r\n new lines
      TextReplaceRE(&wszOutput, &nOutputLen, L"\\n", -1, L"\r\n", -1, FALSE);

      //Include NULL symbol for TextReplace
      ++nOutputLen;

      //Replace to <INPUTFILE>
      if (TextReplace(&wszOutput, &nOutputLen, L"%f", -1, L"<INPUTFILE>", -1, FALSE))
        dwDetectState=DSTATE_SINGLEFILESINGLEPASS;

      //Exclude NULL symbol
      --nOutputLen;
    }
    else if (dwScriptType & STYPE_AVS)
    {
      //Detect first input video file
      pe.lpREGroupStack=0;
      pe.wpStr=wszOutput;
      pe.wpMaxStr=wszOutput + nOutputLen;
      pe.wpPat=L"^\\s*#|^.*(AVISource|DirectShowSource|FFVideoSource|FFmpegSource2)\\(\"(.*)\"";
      pe.wpMaxPat=pe.wpPat + xstrlenW(pe.wpPat);
      pe.dwOptions=REPE_MULTILINE;
      pe.wpDelim=NULL;
      pe.lpCallback=NULL;

      while (PatStructExec(&pe))
      {
        if ((lpREGroupItem=PatGetGroup(pe.lpREGroupStack, 2)) && lpREGroupItem->nStrLen > 0 && lpREGroupItem->nStrLen < MAX_PATH)
        {
          xstrcpynW(wszBuf, lpREGroupItem->wpStrStart, (lpREGroupItem->wpStrEnd - lpREGroupItem->wpStrStart) + 1);
          if (lpREGroupItem=PatGetGroup(pe.lpREGroupStack, 1))
            xstrcpynW(wszMethod, lpREGroupItem->wpStrStart, (lpREGroupItem->wpStrEnd - lpREGroupItem->wpStrStart) + 1);
          bVideoFile=FALSE;

          if (!xstrcmpiW(wszBuf, L"<INPUTFILE>"))
          {
            bVideoFile=TRUE;
          }
          else if (!xstrcmpiW(wszMethod, L"AVISource") ||
                   !xstrcmpiW(wszMethod, L"FFVideoSource"))
          {
            bVideoFile=TRUE;
          }
          else if (wpExt=GetFileExt(wszBuf, -1))
          {
            xprintfW(wszExt, L"|%s|", wpExt);
            if (xstrstrW(L"|mpg|vob|m2v|mpeg|mpv|vdr|avi|divx|m1v|dat|anim|flm|midx|anm|asf|bik|dts|dxa|flv|fli|flc|flx|h261|h263|h264|m4v|mkv|mjp|mlp|mov|mp4|3gp|3g2|mj2|mvi|pmp|rm|rmvb|rpl|smk|swf|vc1|wmv|ts|mts|m2ts|m2t|mxf|ogm|qt|tp|dvr-ms|amv|", -1, wszExt, -1, FALSE, NULL, NULL))
            {
              bVideoFile=TRUE;
            }
          }

          if (bVideoFile)
          {
            if (!(dwDetectState & DSTATE_SINGLEFILESINGLEPASS))
            {
              xstrcpynW(wszInputFile, wszBuf, MAX_PATH);
              dwDetectState=DSTATE_SINGLEFILESINGLEPASS;
            }
            else
            {
              if (!xstrcmpiW(wszInputFile, wszBuf))
              {
                dwDetectState|=DSTATE_SINGLEFILEMULTIPASS;
              }
              else
              {
                dwDetectState&=~DSTATE_SINGLEFILEMULTIPASS;
                dwDetectState|=DSTATE_MULTIFILES;
                break;
              }
            }
          }
        }
        pe.wpStr=pe.lpREGroupStack->first->wpStrEnd;
        if (*(pe.wpStr - 1) == L'\n' || *(pe.wpStr - 1) == L'\r')
          pe.lpREGroupStack->dwOptions&=~REO_NOSTARTLINEBEGIN;
        else
          pe.lpREGroupStack->dwOptions|=REO_NOSTARTLINEBEGIN;
        PatReset(pe.lpREGroupStack);
      }
      PatStructFree(&pe);
    }

    if (dwDetectState & DSTATE_SINGLEFILESINGLEPASS)
    {
      if (dwDetectState & DSTATE_MULTIFILES)
      {
        if (!bNoMessageBox && MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_MULTIFILESFIRST), STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON2|MB_ICONEXCLAMATION) == IDNO)
          goto EndJob;
      }
      if (dwDetectState & DSTATE_SINGLEFILEMULTIPASS)
      {
        xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_MULTIPASS), GetFileName(wszInputFile, -1));
        if (!bNoMessageBox && MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON2|MB_ICONEXCLAMATION) == IDNO)
          goto EndJob;
      }
      if (bSkipParse || !xstrcmpiW(wszInputFile, L"<INPUTFILE>"))
      {
        nOutputLen=nInputLen;
        if (wszOutput=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nOutputLen + 1) * sizeof(wchar_t)))
          xstrcpynW(wszOutput, wpInput, nOutputLen + 1);
        goto EndJob;
      }

      //Include NULL symbol for TextReplace
      ++nOutputLen;

      //Replace to <INPUTFILE>
      TextReplace(&wszOutput, &nOutputLen, wszInputFile, -1, L"<INPUTFILE>", -1, FALSE);

      //Replace to <INPUTFILENOEXT>
      if (wpExt=GetFileExt(wszInputFile, -1))
      {
        xstrcpynW(wszFileNoExt, wszInputFile, wpExt - wszInputFile);
        TextReplace(&wszOutput, &nOutputLen, wszFileNoExt, -1, L"<INPUTFILENOEXT>", -1, FALSE);
      }

      //Exclude NULL symbol
      --nOutputLen;
    }
    else
    {
      if (!bNoMessageBox) MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_NOINPUTFILE), STR_DIALOGTITLE, MB_OK|MB_ICONEXCLAMATION);
    }
  }
  else if (dwScriptType & STYPE_VIRTUALDUB)
  {
    //Detect main input file
    pe.lpREGroupStack=0;
    pe.wpStr=wpInput;
    pe.wpMaxStr=wpInput + nInputLen;

    pe.wpPat=L"(// \\$job.*)?VirtualDub\\.Open\\(\"(.*)\"";
    pe.wpMaxPat=pe.wpPat + xstrlenW(pe.wpPat);
    pe.dwOptions=REPE_MULTILINE;
    pe.wpDelim=NULL;
    pe.lpCallback=NULL;

    while (PatStructExec(&pe))
    {
      if ((lpREGroupItem=PatGetGroup(pe.lpREGroupStack, 2)) && lpREGroupItem->nStrLen > 0 && lpREGroupItem->nStrLen < MAX_PATH)
      {
        xstrcpynW(wszBuf, lpREGroupItem->wpStrStart, (lpREGroupItem->wpStrEnd - lpREGroupItem->wpStrStart) + 1);

        if (!(dwDetectState & DSTATE_SINGLEFILESINGLEPASS))
        {
          xstrcpynW(wszInputFile, wszBuf, MAX_PATH);
          wpScriptStart=pe.lpREGroupStack->first->wpStrStart;
          dwDetectState|=DSTATE_SINGLEFILESINGLEPASS;
          nJobCount=1;
        }
        else
        {
          //Compare file without extension
          if (!xstrcmpinW(wszInputFile, wszBuf, (GetFileExt(wszInputFile, -1) - 1) - wszInputFile))
          {
            dwDetectState|=DSTATE_SINGLEFILEMULTIPASS;
            ++nJobCount;
          }
          else
          {
            xstrcpynW(wszInputFile, wszBuf, MAX_PATH);
            wpScriptStart=pe.lpREGroupStack->first->wpStrStart;
            dwDetectState&=~DSTATE_SINGLEFILEMULTIPASS;
            dwDetectState|=DSTATE_MULTIFILES;
            nJobCount=1;
          }
        }
      }
      else break;

      pe.wpStr=pe.lpREGroupStack->first->wpStrEnd;
    }
    PatStructFree(&pe);

    if (dwDetectState & DSTATE_SINGLEFILESINGLEPASS)
    {
      if (dwDetectState & DSTATE_MULTIFILES)
      {
        if (!bNoMessageBox && MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_MULTIFILESLAST), STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON2|MB_ICONEXCLAMATION) == IDNO)
          goto EndJob;
      }
      if (dwDetectState & DSTATE_SINGLEFILEMULTIPASS)
      {
        xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_MULTIPASS), GetFileName(wszInputFile, -1));
        if (!bNoMessageBox && MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON2|MB_ICONEXCLAMATION) == IDNO)
          goto EndJob;
      }
      if (bSkipParse || !xstrcmpiW(wszInputFile, L"<INPUTFILE>"))
      {
        nOutputLen=nInputLen;
        if (wszOutput=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nOutputLen + 1) * sizeof(wchar_t)))
          xstrcpynW(wszOutput, wpInput, nOutputLen + 1);
        if (!xstrstrW(wszOutput, nOutputLen, L"<OUTPUTFILE>", -1, TRUE, NULL, NULL))
          dwDetectState|=DSTATE_NOOUTPUTFILE;
        goto EndJob;
      }

      nOutputLen=(wpInput + nInputLen) - wpScriptStart;
      if (wszOutput=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nOutputLen + 1) * sizeof(wchar_t)))
        xstrcpynW(wszOutput, wpScriptStart, nOutputLen + 1);

      //Replace to <INPUTFILENUMBER>
      if (wszOutput)
      {
        TextReplaceRE(&wszOutput, &nOutputLen, L"(// \\$job )\".*\"", -1, L"\\1\"File <INPUTFILENUMBER>\"", -1, FALSE);
      }

      //Include NULL symbol for TextReplace
      ++nOutputLen;

      //Replace to <INPUTFILENOSLASH>
      xstrrepW(wszInputFile, -1, L"\\\\", 2, L"\\", 1, TRUE, wszFileNoSlash, NULL);
      TextReplace(&wszOutput, &nOutputLen, wszFileNoSlash, -1, L"<INPUTFILENOSLASH>", -1, FALSE);

      //Replace to <INPUTFILE>
      TextReplace(&wszOutput, &nOutputLen, wszInputFile, -1, L"<INPUTFILE>", -1, FALSE);

      //Replace to <INPUTFILENOEXT>
      if (wpExt=GetFileExt(wszInputFile, -1))
      {
        xstrcpynW(wszFileNoExt, wszInputFile, wpExt - wszInputFile);
        TextReplace(&wszOutput, &nOutputLen, wszFileNoExt, -1, L"<INPUTFILENOEXT>", -1, FALSE);
      }

      //Exclude NULL symbol
      --nOutputLen;

      //Replace to <OUTPUTFILENOSLASH>
      if (wszOutput)
      {
        TextReplaceRE(&wszOutput, &nOutputLen, L"(// \\$output )\"[^<].*\"", -1, L"\\1\"<OUTPUTFILENOSLASH>\"", -1, FALSE);
      }

      if (wszOutput)
      {
        //Detect output directory
        bCheckWAV=FALSE;

        while (!*wszOutputDir)
        {
          pe.lpREGroupStack=0;
          pe.wpStr=wszOutput;
          pe.wpMaxStr=wszOutput + nOutputLen;
          if (bCheckWAV)
            pe.wpPat=L"VirtualDub\\.SaveWAV\\(\"([^<].*)\"";
          else
            pe.wpPat=L"VirtualDub\\.SaveAVI\\(\"([^<].*)\"";
          pe.wpMaxPat=pe.wpPat + xstrlenW(pe.wpPat);
          pe.dwOptions=REPE_MULTILINE;
          pe.wpDelim=NULL;
          pe.lpCallback=NULL;

          while (PatStructExec(&pe))
          {
            if ((lpREGroupItem=PatGetGroup(pe.lpREGroupStack, 1)) && lpREGroupItem->nStrLen > 0 && lpREGroupItem->nStrLen < MAX_PATH)
            {
              xstrcpynW(wszBuf, lpREGroupItem->wpStrStart, (lpREGroupItem->wpStrEnd - lpREGroupItem->wpStrStart) + 1);
              xstrrepW(wszBuf, -1, L"\\\\", 2, L"\\", 1, TRUE, wszFileNoSlash, NULL);
              GetFileDir(wszFileNoSlash, -1, wszOutputDir, MAX_PATH);
            }
            else break;

            pe.wpStr=pe.lpREGroupStack->first->wpStrEnd;
          }
          PatStructFree(&pe);

          if (bCheckWAV) break;
          bCheckWAV=TRUE;
        }

        //Replace to <OUTPUTFILE>
        if (!(nReplaceCount=TextReplaceRE(&wszOutput, &nOutputLen, L"(VirtualDub\\.SaveAVI\\()\"[^<].*\"", -1, L"\\1\"<OUTPUTFILE>\"", -1, FALSE)))
          nReplaceCount=TextReplaceRE(&wszOutput, &nOutputLen, L"(VirtualDub\\.SaveWAV\\()\"[^<].*\"", -1, L"\\1\"<OUTPUTFILE>\"", -1, FALSE);

        if (!nReplaceCount)
        {
          //Output file name is based on input file name
          dwDetectState|=DSTATE_NOOUTPUTFILE;
        }
      }

      //Remove "VirtualDub.subset"
      if (wszOutput)
      {
        if (xstrstrW(wszOutput, nOutputLen, L"VirtualDub.subset", -1, FALSE, NULL, NULL))
        {
          if (!bNoMessageBox && MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_REMOVESUBSET), STR_DIALOGTITLE, MB_YESNO|MB_DEFBUTTON1|MB_ICONQUESTION) == IDNO)
          {
            //Don't remove "subset"
          }
          else TextReplaceRE(&wszOutput, &nOutputLen, L"^(VirtualDub\\.subset.*$)(\\n?)", -1, L"//\\1\\2", -1, FALSE);
        }
      }

      //Remove "// $done"
      if (wszOutput)
      {
        TextReplaceRE(&wszOutput, &nOutputLen, L"// \\$done\\n*", -1, L"", -1, FALSE);
      }
    }
    else
    {
      if (!bNoMessageBox && MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_ADDOPENSAVECOMMANDS), STR_DIALOGTITLE, MB_YESNO|MB_ICONQUESTION) == IDNO)
        goto EndJob;

      for (;;)
      {
        nTextLen=xprintfW(wszOutput, L"\
// $job \"File <INPUTFILENUMBER>\"\r\n\
// $input \"<INPUTFILE>\"\r\n\
// $output \"<OUTPUTFILE>\"\r\n\
// $state 0\r\n\
// $start_time 0 0\r\n\
// $end_time 0 0\r\n\
// $script\r\n\
\r\n\
VirtualDub.Open(\"<INPUTFILE>\",\"\",0);\r\n\
%s\r\n\
VirtualDub.SaveAVI(\"<OUTPUTFILE>\");\r\n\
VirtualDub.Close();\r\n\
\r\n\
// $endjob\r\n\
//\r\n\
//--------------------------------------------------\r\n\
", wpScriptStart);
        if (!wszOutput && (wszOutput=(wchar_t *)GlobalAlloc(GMEM_FIXED, nTextLen * sizeof(wchar_t))))
        {
          nOutputLen=nTextLen;
          nJobCount=1;
          continue;
        }
        break;
      }
      dwDetectState|=DSTATE_SINGLEFILESINGLEPASS;
    }
  }

  EndJob:
  *wppOutput=wszOutput;
  *lpnOutputLen=nOutputLen;

  if (bOutputDirEmpty && *wszOutputDir)
    SetDlgItemTextWide(hDlg, IDC_OUTPUTDIR, wszOutputDir);
}

int SaveScript(HWND hDlg, HWND hWndList, const wchar_t *wpOutputScriptFile)
{
  LVITEMW lvi;
  wchar_t wszInputFile[MAX_PATH];
  wchar_t wszOutputFile[MAX_PATH];
  wchar_t wszOutputFileDir[MAX_PATH];
  wchar_t wszOutputFileName[MAX_PATH];
  wchar_t wszFileNoExt[MAX_PATH];
  wchar_t wszFileSlashes[MAX_PATH];
  wchar_t *wszScript=NULL;
  char *szScript;
  HANDLE hFile=NULL;
  DWORD dwBytesWritten;
  INT_PTR nScriptLen;
  int nFilesCount=(int)SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0);
  int nIndex=0;

  if (!nFilesCount)
  {
    if (!bNoMessageBox) MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_NOFILESFORPROCESS), STR_DIALOGTITLE, MB_OK|MB_ICONERROR);
    return -1;
  }

  if (dwScriptType & STYPE_VIRTUALDUB)
  {
    if ((hFile=CreateFileWide(wpOutputScriptFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
    {
      xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_CANNOTCREATEFILE), wpOutputScriptFile);
      if (!bNoMessageBox) MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_OK|MB_ICONERROR);
      return -1;
    }

    //Write $numjobs
    nScriptLen=xprintfW(wszBuf, L"\
// $numjobs %d\r\n\
//\r\n\
\r\n\
", nFilesCount * nJobCount, wszTemplate);
    if (szScript=AllocAnsiLen(wszBuf, nScriptLen + 1))
    {
      WriteFile(hFile, szScript, lstrlenA(szScript), &dwBytesWritten, NULL);
      GlobalFree((HGLOBAL)szScript);
    }
  }

  //Write jobs
  for (;;)
  {
    lvi.mask=LVIF_TEXT;
    lvi.pszText=wszInputFile;
    lvi.cchTextMax=MAX_PATH;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVSI_INPUTFILE;
    if (!ListView_GetItemWide(hWndList, &lvi))
      break;

    lvi.mask=LVIF_TEXT;
    lvi.pszText=wszOutputFileName;
    lvi.cchTextMax=MAX_PATH;
    lvi.iItem=nIndex;
    lvi.iSubItem=LVSI_OUTPUTFILENAME;
    if (!ListView_GetItemWide(hWndList, &lvi))
      break;

    if (*wszOutputDir)
      xstrcpynW(wszOutputFileDir, wszOutputDir, MAX_PATH);
    else
      GetFileDir(wszInputFile, -1, wszOutputFileDir, MAX_PATH);

    if (wszScript=(wchar_t *)GlobalAlloc(GMEM_FIXED, (nTemplateLen + 1) * sizeof(wchar_t)))
      nScriptLen=xstrcpynW(wszScript, wszTemplate, nTemplateLen + 1);

    //Include NULL symbol for TextReplace
    ++nScriptLen;

    //Replace from <INPUTFILENUMBER>
    xprintfW(wszBuf, L"%d", nIndex + 1);
    TextReplace(&wszScript, &nScriptLen, L"<INPUTFILENUMBER>", -1, wszBuf, -1, FALSE);

    ////Input
    if (dwScriptType & STYPE_VIRTUALDUB)
      xstrrepW(wszInputFile, -1, L"\\", 1, L"\\\\", 2, TRUE, wszFileSlashes, NULL);
    else
      xstrcpynW(wszFileSlashes, wszInputFile, MAX_PATH);

    //Replace from <INPUTFILENOSLASH>
    if (dwScriptType & STYPE_VIRTUALDUB)
      TextReplace(&wszScript, &nScriptLen, L"<INPUTFILENOSLASH>", -1, wszInputFile, -1, FALSE);

    //Replace from <INPUTFILENOEXT>
    xstrcpynW(wszFileNoExt, wszFileSlashes, GetFileExt(wszFileSlashes, -1) - wszFileSlashes);
    TextReplace(&wszScript, &nScriptLen, L"<INPUTFILENOEXT>", -1, wszFileNoExt, -1, FALSE);

    //Replace from <INPUTFILE>
    TextReplace(&wszScript, &nScriptLen, L"<INPUTFILE>", -1, wszFileSlashes, -1, FALSE);

    ////Output
    xprintfW(wszOutputFile, L"%s\\%s", wszOutputFileDir, wszOutputFileName);
    if (dwScriptType & STYPE_VIRTUALDUB)
      xstrrepW(wszOutputFile, -1, L"\\", 1, L"\\\\", 2, TRUE, wszFileSlashes, NULL);
    else
      xstrcpynW(wszFileSlashes, wszOutputFile, MAX_PATH);

    //Replace from <OUTPUTFILENOSLASH>
    if (dwScriptType & STYPE_VIRTUALDUB)
      TextReplace(&wszScript, &nScriptLen, L"<OUTPUTFILENOSLASH>", -1, wszOutputFile, -1, FALSE);

    //Replace from <OUTPUTFILENOEXT>
    xstrcpynW(wszFileNoExt, wszFileSlashes, GetFileExt(wszFileSlashes, -1) - wszFileSlashes);
    TextReplace(&wszScript, &nScriptLen, L"<OUTPUTFILENOEXT>", -1, wszFileNoExt, -1, FALSE);

    //Replace from <OUTPUTFILE>
    TextReplace(&wszScript, &nScriptLen, L"<OUTPUTFILE>", -1, wszFileSlashes, -1, FALSE);

    //Create AVS file
    if (dwScriptType & STYPE_AVS)
    {
      if ((hFile=CreateFileWide(wszOutputFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
      {
        xprintfW(wszBuf, GetLangStringW(wLangSystem, STRID_CANNOTCREATEFILE), wszOutputFile);
        if (!bNoMessageBox) MessageBoxW(hDlg, wszBuf, STR_DIALOGTITLE, MB_OK|MB_ICONERROR);
        return -1;
      }
    }

    //Write to file
    if (szScript=AllocAnsiLen(wszScript, nScriptLen))
    {
      if (!WriteFile(hFile, szScript, lstrlenA(szScript), &dwBytesWritten, NULL))
        break;
      GlobalFree((HGLOBAL)szScript);
    }
    GlobalFree((HGLOBAL)wszScript);
    wszScript=NULL;

    //Close AVS file
    if (dwScriptType & STYPE_AVS)
    {
      CloseHandle(hFile);
    }
    ++nIndex;
  }

  if (dwScriptType & STYPE_VIRTUALDUB)
  {
    //Write $done
    WriteFile(hFile, "// $done\r\n", lstrlenA("// $done\r\n"), &dwBytesWritten, NULL);

    CloseHandle(hFile);
  }
  return nIndex;
}

int TextReplace(wchar_t **wppText, INT_PTR *lpnTextLen, const wchar_t *wpIt, int nItLen, const wchar_t *wpWith, int nWithLen, BOOL bSensitive)
{
  wchar_t *wszText;
  INT_PTR nTextLen;
  int nChanges;

  if (nChanges=xstrrepW(*wppText, *lpnTextLen, wpIt, nItLen, wpWith, nWithLen, bSensitive, NULL, &nTextLen))
  {
    if (wszText=(wchar_t *)GlobalAlloc(GMEM_FIXED, nTextLen * sizeof(wchar_t)))
    {
      xstrrepW(*wppText, *lpnTextLen, wpIt, nItLen, wpWith, nWithLen, bSensitive, wszText, NULL);
      GlobalFree((HGLOBAL)*wppText);
      *wppText=wszText;
      *lpnTextLen=nTextLen;
    }
  }
  return nChanges;
}

int TextReplaceRE(wchar_t **wppText, INT_PTR *lpnTextLen, const wchar_t *wpIt, int nItLen, const wchar_t *wpWith, int nWithLen, BOOL bSensitive)
{
  PATREPLACE pr;
  INT_PTR nResultTextLen;

  if (nItLen == -1)
    nItLen=xstrlenW(wpIt);
  if (nWithLen == -1)
    nWithLen=xstrlenW(wpWith);

  pr.wpStr=*wppText;
  pr.wpMaxStr=*wppText + *lpnTextLen;
  pr.wpPat=wpIt;
  pr.wpMaxPat=pr.wpPat + nItLen;
  pr.wpRep=wpWith;
  pr.wpMaxRep=pr.wpRep + nWithLen;
  pr.dwOptions=(bSensitive?REPE_MATCHCASE:0)|REPE_GLOBAL|REPE_MULTILINE;
  pr.wpDelim=NULL;
  pr.wpNewLine=NULL;
  pr.wszResult=NULL;
  nResultTextLen=PatReplace(&pr);

  if (pr.nReplaceCount)
  {
    if (pr.wszResult=(wchar_t *)GlobalAlloc(GMEM_FIXED, nResultTextLen * sizeof(wchar_t)))
    {
      nResultTextLen=PatReplace(&pr);
      GlobalFree((HGLOBAL)*wppText);
      *wppText=pr.wszResult;
      *lpnTextLen=nResultTextLen;
    }
  }
  return pr.nReplaceCount;
}

int GetExeDir(HINSTANCE hInstance, wchar_t *wszExeDir, int nLen)
{
  if (nLen=GetModuleFileNameWide(hInstance, wszExeDir, nLen))
  {
    while (nLen > 0 && wszExeDir[nLen] != '\\') --nLen;
    wszExeDir[nLen]='\0';
  }
  return nLen;
}

int GetFileDir(const wchar_t *wpFile, int nFileLen, wchar_t *wszFileDir, DWORD dwFileDirLen)
{
  const wchar_t *wpCount;

  if (nFileLen == -1) nFileLen=(int)xstrlenW(wpFile);
  if (wszFileDir) wszFileDir[0]=L'\0';

  for (wpCount=wpFile + nFileLen - 1; wpCount >= wpFile; --wpCount)
  {
    if (*wpCount == L'\\')
      return (int)xstrcpynW(wszFileDir, wpFile, min(dwFileDirLen, (DWORD)(wpCount - wpFile) + 1));
  }
  return 0;
}

const wchar_t* GetFileName(const wchar_t *wpFile, int nFileLen)
{
  const wchar_t *wpCount;

  if (nFileLen == -1) nFileLen=(int)xstrlenW(wpFile);

  for (wpCount=wpFile + nFileLen - 1; wpCount >= wpFile; --wpCount)
  {
    if (*wpCount == L'\\')
      return wpCount + 1;
  }
  return wpFile;
}

int GetBaseName(const wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen)
{
  int nFileLen=(int)xstrlenW(wpFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (wpFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (wpFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  xstrcpynW(wszBaseName, wpFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

const wchar_t* GetFileExt(const wchar_t *wpFile, int nFileLen)
{
  const wchar_t *wpCount;

  if (nFileLen == -1) nFileLen=(int)xstrlenW(wpFile);

  for (wpCount=wpFile + nFileLen - 1; wpCount >= wpFile; --wpCount)
  {
    if (*wpCount == L'.') return wpCount + 1;
    if (*wpCount == L'\\') break;
  }
  return NULL;
}

int IsFile(const wchar_t *wpFile)
{
  DWORD dwAttr;

  dwAttr=GetFileAttributesWide(wpFile);
  if (dwAttr == INVALID_FILE_ATTRIBUTES)
    return ERROR_INVALID_HANDLE;
  if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
    return ERROR_DIRECTORY;
  return ERROR_SUCCESS;
}

BOOL TrimTrailBackSlash(wchar_t *wszDir)
{
  int nLen;
  int i;

  nLen=(int)xstrlenW(wszDir);

  for (i=nLen - 1; i > 0 && wszDir[i] == '\\'; --i)
  {
    wszDir[i]=L'\0';
  }
  if (i < nLen - 1) return TRUE;
  return FALSE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

wchar_t* GetCommandLineParamsWide()
{
  if (WideGlobal_bOldWindows)
    MultiByteToWideChar(CP_ACP, 0, GetCommandLineParamsA(), -1, wszCmdLine, COMMANDLINE_SIZE);
  else
    xstrcpynW(wszCmdLine, GetCommandLineParamsW(), COMMANDLINE_SIZE);
  return wszCmdLine;
}

char* GetCommandLineParamsA()
{
  char *lpCmdLine=GetCommandLineA();

  if (*lpCmdLine++ == '\"')
    while (*lpCmdLine != '\"' && *lpCmdLine != '\0') ++lpCmdLine;
  else
    while (*lpCmdLine != ' ' && *lpCmdLine != '\0') ++lpCmdLine;
  if (*lpCmdLine != '\0')
    while (*++lpCmdLine == ' ');

  return lpCmdLine;
}

wchar_t* GetCommandLineParamsW()
{
  wchar_t *lpwCmdLine=GetCommandLineW();

  if (*lpwCmdLine++ == '\"')
    while (*lpwCmdLine != '\"' && *lpwCmdLine != '\0') ++lpwCmdLine;
  else
    while (*lpwCmdLine != ' ' && *lpwCmdLine != '\0') ++lpwCmdLine;
  if (*lpwCmdLine != '\0')
    while (*++lpwCmdLine == ' ');

  return lpwCmdLine;
}

int GetCommandLineArg(const wchar_t *wpCmdLine, wchar_t *wszArg, int nArgMax, const wchar_t **wpNextArg)
{
  const wchar_t *wpCount=wpCmdLine;
  wchar_t *wpArgSet=wszArg;
  wchar_t *wpArgSetMax=wszArg + nArgMax - 1;
  wchar_t wchInitStopChar;
  wchar_t wchCurStopChar;

  while (*wpCount == L' ') ++wpCount;

  if (*wpCount == L'\"' || *wpCount == L'\'' || *wpCount == L'`')
    wchInitStopChar=*wpCount++;
  else
    wchInitStopChar=L' ';
  wchCurStopChar=wchInitStopChar;

  for (; *wpCount; ++wpCount)
  {
    if (wchCurStopChar == L' ')
    {
      if (*wpCount == L' ')
        break;
      if (*wpCount == L'\"' || *wpCount == L'\'' || *wpCount == L'`')
        wchCurStopChar=*wpCount;
    }
    else if (wchCurStopChar == *wpCount)
    {
      if (wchCurStopChar == wchInitStopChar)
        break;
      wchCurStopChar=L' ';
    }

    if (wpArgSet < wpArgSetMax)
    {
      if (wszArg) *wpArgSet=*wpCount;
      ++wpArgSet;
    }
  }
  if (wszArg) *wpArgSet=L'\0';

  if (wpNextArg)
  {
    if (*wpCount) ++wpCount;
    for (*wpNextArg=wpCount; **wpNextArg == L' '; ++*wpNextArg);
  }
  return (int)(wpArgSet - wszArg);
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_HELP)
      return L"\
|   BatchWizardMod.exe [\x043F\x0430\x0440\x0430\x043C\x0435\x0442\x0440\x044B] \"file1.ext\" \"file2.ext\"\n\
|\n\
|   \x041F\x0435\x0440\x0430\x043C\x0435\x0442\x0440\x044B\x0020\x043A\x043E\x043C\x0430\x043D\x0434\x043D\x043E\x0439\x0020\x0441\x0442\x0440\x043E\x043A\x0438:\n\
|\n\
|   /S=\"Script.vcf\"\n\
|     \x0424\x0430\x0439\x043B\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0430.\n\
|\n\
|   /P\n\
|     \x041E\x0431\x0440\x0430\x0431\x043E\x0442\x0430\x0442\x044C\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0020\x0028\x0433\x0435\x043D\x0435\x0440\x0438\x0440\x043E\x0432\x0430\x0442\x044C\x0020\x0448\x0430\x0431\x043B\x043E\x043D\x0029.\n\
|\n\
|   /O=\"C:\\Output\"\n\
|     \x0414\x0438\x0440\x0435\x043A\x0442\x043E\x0440\x0438\x044F\x0020\x0438\x0441\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432.\n\
|\n\
|   /FF=\"\x041D\x0430\x0439\x0442\x0438\x0020\x0447\x0442\x043E\"\n\
|     \x0424\x0438\x043B\x044C\x0442\x0440\x003A\x0020\x043D\x0430\x0439\x0442\x0438.\n\
|\n\
|   /FR=\"\x0417\x0430\x043C\x0435\x043D\x0438\x0442\x044C\x0020\x0447\x0435\x043C\"\n\
|     \x0424\x0438\x043B\x044C\x0442\x0440\x003A\x0020\x0437\x0430\x043C\x0435\x043D\x0438\x0442\x044C.\n\
|\n\
|   /FC\n\
|     \x0424\x0438\x043B\x044C\x0442\x0440\x003A\x0020\x0443\x0447\x0438\x0442\x044B\x0432\x0430\x0442\x044C\x0020\x0440\x0435\x0433\x0438\x0441\x0442\x0440.\n\
|\n\
|   /FE\n\
|     \x0424\x0438\x043B\x044C\x0442\x0440\x003A\x0020\x0440\x0435\x0433\x0443\x043B\x044F\x0440\x043D\x044B\x0435\x0020\x0432\x044B\x0440\x0430\x0436\x0435\x043D\x0438\x044F.\n\
|\n\
|   /SAVE=\"Script.jobs\"\n\
|     \x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0020\x0434\x043B\x044F VirtualDub'\x0430.\n\
|\n\
|   /EXEC\n\
|     \x0414\x043B\x044F VirtualDub - \x0437\x0430\x043F\x0443\x0441\x0442\x0438\x0442\x044C\x0020\x0441\x0433\x0435\x043D\x0435\x0440\x0438\x0440\x043E\x0432\x0430\x043D\x043D\x044B\x0439\x0020\x0441\x043A\x0440\x0438\x043F\x0442.\n\
|     \x0414\x043B\x044F AviSynth - \x0441\x043E\x0437\x0434\x0430\x0442\x044C .avs.\n\
|\n\
|   /NOMSG\n\
|     \x0411\x0435\x0437\x0020\x0437\x0430\x043F\x0440\x043E\x0441\x043E\x0432.\n\
|\n\
|   /QUIT\n\
|     \x0412\x044B\x0439\x0442\x0438\x0020\x0438\x0437\x0020\x043F\x0440\x043E\x0433\x0440\x0430\x043C\x043C\x044B.\n\
\n";
    if (nStringID == STRID_OUTPUTDIRNOTFOUND)
      return L"\x0414\x0438\x0440\x0435\x043A\x0442\x043E\x0440\x0438\x044F\x0020\x0438\x0441\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432 \"%s\" \x043D\x0435\x0020\x043D\x0430\x0439\x0434\x0435\x043D\x0430\x002E";
    if (nStringID == STRID_OUTPUTDIREMPTY)
      return L"\x0414\x0438\x0440\x0435\x043A\x0442\x043E\x0440\x0438\x044F\x0020\x0438\x0441\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x0020\x043D\x0435\x0020\x0437\x0430\x0434\x0430\x043D\x0430\x002E\x0020\x0421\x043E\x0437\x0434\x0430\x0442\x044C\x0020\x0061\x0076\x0073\x0020\x043F\x043E\x0020\x043C\x0435\x0441\x0442\x0443\x0020\x0440\x0430\x0441\x043F\x043E\x043B\x043E\x0436\x0435\x043D\x0438\x044F\x0020\x0432\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x003F";
    if (nStringID == STRID_OUTPUTDIRNOTFOUNDCONTINUE)
      return L"\x0414\x0438\x0440\x0435\x043A\x0442\x043E\x0440\x0438\x044F\x0020\x0438\x0441\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432 \"%s\" \x043D\x0435\x0020\x043D\x0430\x0439\x0434\x0435\x043D\x0430\x002E\x0020\x041F\x0440\x043E\x0434\x043E\x043B\x0436\x0438\x0442\x044C\x003F";
    if (nStringID == STRID_VIRTUALDUBNOTFOUND)
      return L"\x0418\x0441\x043F\x043E\x043B\x043D\x044F\x0435\x043C\x044B\x0439\x0020\x0444\x0430\x0439\x043B\x0020\x0056\x0069\x0072\x0074\x0075\x0061\x006C\x0044\x0075\x0062\x0020\x043D\x0435\x0020\x043D\x0430\x0439\x0434\x0435\x043D\x002E";
    if (nStringID == STRID_MULTIFILESFIRST)
      return L"\x0421\x043A\x0440\x0438\x043F\x0442\x0020\x0441\x043E\x0434\x0435\x0440\x0436\x0438\x0442\x0020\x043C\x043D\x043E\x0436\x0435\x0441\x0442\x0432\x043E\x0020\x0432\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x002E\x0020\x0411\x0443\x0434\x0435\x0442\x0020\x043E\x0431\x0440\x0430\x0431\x043E\x0442\x0430\x043D\x0020\x0442\x043E\x043B\x044C\x043A\x043E\x0020\x043F\x0435\x0440\x0432\x044B\x0439\x0020\x0432\x0445\x043E\x0434\x044F\x0449\x0438\x0439\x0020\x0444\x0430\x0439\x043B\x002E\x0020\x041F\x0440\x043E\x0434\x043E\x043B\x0436\x0438\x0442\x044C\x003F";
    if (nStringID == STRID_MULTIFILESLAST)
      return L"\x0421\x043A\x0440\x0438\x043F\x0442\x0020\x0441\x043E\x0434\x0435\x0440\x0436\x0438\x0442\x0020\x043C\x043D\x043E\x0436\x0435\x0441\x0442\x0432\x043E\x0020\x0432\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x002E\x0020\x0411\x0443\x0434\x0443\x0442\x0020\x0438\x043C\x043F\x043E\x0440\x0442\x0438\x0440\x043E\x0432\x0430\x043D\x044B\x0020\x043D\x0430\x0441\x0442\x0440\x043E\x0439\x043A\x0438\x0020\x043F\x043E\x0441\x043B\x0435\x0434\x043D\x0435\x0433\x043E\x0020\x0432\x0445\x043E\x0434\x044F\x0449\x0435\x0433\x043E\x0020\x0444\x0430\x0439\x043B\x0430\x002E\x0020\x041F\x0440\x043E\x0434\x043E\x043B\x0436\x0438\x0442\x044C\x003F";
    if (nStringID == STRID_MULTIPASS)
      return L"\x0424\x0430\x0439\x043B \"%s\" \x0432\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0435\x0020\x043E\x0431\x0440\x0430\x0431\x0430\x0442\x044B\x0432\x0430\x0435\x0442\x0441\x044F\x0020\x043D\x0435\x0441\x043A\x043E\x043B\x044C\x043A\x043E\x0020\x0440\x0430\x0437\x002E\x0020\x0412\x044B\x0020\x0443\x0432\x0435\x0440\x0435\x043D\x044B\x003F";
    if (nStringID == STRID_ADDOPENSAVECOMMANDS)
      return L"\x0412\x0445\x043E\x0434\x044F\x0449\x0438\x0439\x0020\x0444\x0430\x0439\x043B\x0020\x043D\x0435\x0020\x043E\x0431\x043D\x0430\x0440\x0443\x0436\x0435\x043D\x002E\x0020\x0414\x043E\x0431\x0430\x0432\x0438\x0442\x044C\x0020\x043A\x043E\x043C\x0430\x043D\x0434\x044B \"Open\" \x0438 \"SaveAVI\"?";
    if (nStringID == STRID_NOINPUTFILE)
      return L"\x0412\x0445\x043E\x0434\x044F\x0449\x0438\x0439\x0020\x0444\x0430\x0439\x043B\x0020\x043D\x0435\x0020\x043E\x0431\x043D\x0430\x0440\x0443\x0436\x0435\x043D\x002E";
    if (nStringID == STRID_REMOVESUBSET)
      return L"\x041E\x0431\x043D\x0430\x0440\x0443\x0436\x0435\x043D\x0430\x0020\x043A\x043E\x043C\x0430\x043D\x0434\x0430 \"VirtualDub.subset\". \x0417\x0430\x043A\x043E\x043C\x043C\x0435\x043D\x0442\x0438\x0440\x043E\x0432\x0430\x0442\x044C\x0020\x044D\x0442\x0438\x0020\x0441\x0442\x0440\x043E\x043A\x0438\x003F";
    if (nStringID == STRID_NOFILESFORPROCESS)
      return L"\x041D\x0435\x0442\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x0020\x0434\x043B\x044F\x0020\x043E\x0431\x0440\x0430\x0431\x043E\x0442\x043A\x0438\x002E\x0020\x0421\x043F\x0438\x0441\x043E\x043A\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x0020\x043F\x0443\x0441\x0442\x002E";
    if (nStringID == STRID_CANNOTCREATEFILE)
      return L"\x041D\x0435\x0432\x043E\x0437\x043C\x043E\x0436\x043D\x043E\x0020\x0441\x043E\x0437\x0434\x0430\x0442\x044C\x0020\x0444\x0430\x0439\x043B \"%s\".";
    if (nStringID == STRID_DONEFILES)
      return L"\x041E\x0431\x0440\x0430\x0431\x043E\x0442\x0430\x043D\x043E\x0020\x0444\x0430\x0439\x043B\x043E\x0432: %d";
    if (nStringID == STRID_CREATETEMPLATE)
      return L"\x0421\x043E\x0437\x0434\x0430\x0442\x044C\x0020\x0448\x0430\x0431\x043B\x043E\x043D\x0020\x043D\x0430\x0020\x043E\x0441\x043D\x043E\x0432\x0435\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0430";
    if (nStringID == STRID_SCRIPTFILE)
      return L"\x0421\x043A\x0440\x0438\x043F\x0442\x003A";
    if (nStringID == STRID_GENERATE)
      return L"\x0413\x0435\x043D\x0435\x0440\x0438\x0440\x043E\x0432\x0430\x0442\x044C";
    if (nStringID == STRID_LISTLABEL)
      return L"\x0421\x043F\x0438\x0441\x043E\x043A\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x0020\x0434\x043B\x044F\x0020\x043F\x0430\x043A\x0435\x0442\x043D\x043E\x0439\x0020\x043E\x0431\x0440\x0430\x0431\x043E\x0442\x043A\x0438\x0020\x0028\x0434\x043E\x0431\x0430\x0432\x044C\x0442\x0435\x0020\x043C\x0435\x0442\x043E\x0434\x043E\x043C\x0020\x043F\x0435\x0440\x0435\x0442\x0430\x0441\x043A\x0438\x0432\x0430\x043D\x0438\x044F\x0029\x003A";
    if (nStringID == STRID_OUTPUTDIR)
      return L"\x0414\x0438\x0440\x0435\x043A\x0442\x043E\x0440\x0438\x044F\x0020\x0432\x044B\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432\x003A";
    if (nStringID == STRID_OUTPUTFILTER)
      return L"\x0424\x0438\x043B\x044C\x0442\x0440\x0020\x0438\x043C\x0435\x043D\x0020\x0432\x044B\x0445\x043E\x0434\x044F\x0449\x0438\x0445\x0020\x0444\x0430\x0439\x043B\x043E\x0432";
    if (nStringID == STRID_FINDIT)
      return L"\x041D\x0430\x0439\x0442\x0438\x003A";
    if (nStringID == STRID_REPLACEWITH)
      return L"\x0417\x0430\x043C\x0435\x043D\x0438\x0442\x044C\x003A";
    if (nStringID == STRID_MATCHCASE)
      return L"\x0423\x0447\x0438\x0442\x044B\x0432\x0430\x0442\x044C\x0020\x0440\x0435\x0433\x0438\x0441\x0442\x0440";
    if (nStringID == STRID_REGEXP)
      return L"\x0420\x0435\x0433\x0443\x043B\x044F\x0440\x043D\x044B\x0435\x0020\x0432\x044B\x0440\x0430\x0436\x0435\x043D\x0438\x044F";
    if (nStringID == STRID_CREATE)
      return L"\x0421\x043E\x0437\x0434\x0430\x0442\x044C";
    if (nStringID == STRID_EXECUTE)
      return L"\x0412\x044B\x043F\x043E\x043B\x043D\x0438\x0442\x044C";
    if (nStringID == STRID_SAVE)
      return L"\x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C\x002E\x002E\x002E";
    if (nStringID == STRID_INPUTFILE)
      return L"\x0412\x0445\x043E\x0434\x044F\x0449\x0438\x0439\x0020\x0444\x0430\x0439\x043B";
    if (nStringID == STRID_OUTPUTFILENAME)
      return L"\x0418\x043C\x044F\x0020\x0432\x044B\x0445\x043E\x0434\x044F\x0449\x0435\x0433\x043E\x0020\x0444\x0430\x0439\x043B\x0430";
    if (nStringID == STRID_EMPTY)
      return L"\x003C\x041F\x0423\x0421\x0422\x041E\x003E";
    if (nStringID == STRID_TEMPLATE)
      return L"\x0428\x0430\x0431\x043B\x043E\x043D";
    if (nStringID == STRID_SCRIPTTYPE)
      return L"\x0422\x0438\x043F\x0020\x0441\x043A\x0440\x0438\x043F\x0442\x0430";
    if (nStringID == STRID_SCRIPTVIRTUALDUB)
      return L"VirtualDub \x0441\x043A\x0440\x0438\x043F\x0442";
    if (nStringID == STRID_SCRIPTAVS)
      return L"AVS/AVST \x0441\x043A\x0440\x0438\x043F\x0442";
    if (nStringID == STRID_NOPARSE)
      return L"\x041D\x0435\x0020\x043E\x0431\x0440\x0430\x0431\x0430\x0442\x044B\x0432\x0430\x0442\x044C\x0020\x0432\x0432\x0435\x0434\x0435\x043D\x043D\x044B\x0439\x0020\x0442\x0435\x043A\x0441\x0442";
    if (nStringID == STRID_MENU_ITEMDELETE)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C\tDelete";
    if (nStringID == STRID_MENU_ITEMMOVEUP)
      return L"\x041F\x0435\x0440\x0435\x043C\x0435\x0441\x0442\x0438\x0442\x044C\x0020\x0432\x0432\x0435\x0440\x0445\tAlt+Up";
    if (nStringID == STRID_MENU_ITEMMOVEDOWN)
      return L"\x041F\x0435\x0440\x0435\x043C\x0435\x0441\x0442\x0438\x0442\x044C\x0020\x0432\x043D\x0438\x0437\tAlt+Down";
    if (nStringID == STRID_MENU_ITEMCLEAR)
      return L"\x041E\x0447\x0438\x0441\x0442\x0438\x0442\x044C";
    if (nStringID == STRID_CLOSE)
      return L"\x0417\x0430\x043A\x0440\x044B\x0442\x044C";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_HELP)
      return L"\
|   BatchWizardMod.exe [parameters] \"file1.ext\" \"file2.ext\"\n\
|\n\
|   Command line parameters:\n\
|\n\
|   /S=\"Script.vcf\"\n\
|     Script file.\n\
|\n\
|   /P\n\
|     Parse script (generate template).\n\
|\n\
|   /O=\"C:\\Output\"\n\
|     Output directory.\n\
|\n\
|   /FF=\"Find it\"\n\
|     Filter: find it.\n\
|\n\
|   /FR=\"Replace with\"\n\
|     Filter: replace with.\n\
|\n\
|   /FC\n\
|     Filter: case sensitive.\n\
|\n\
|   /FE\n\
|     Filter: regular expressions.\n\
|\n\
|   /SAVE=\"Script.jobs\"\n\
|     Save script for VirtualDub.\n\
|\n\
|   /EXEC\n\
|     For VirtualDub - execute generated script.\n\
|     For AviSynth - creates .avs.\n\
|\n\
|   /NOMSG\n\
|     No message boxes.\n\
|\n\
|   /QUIT\n\
|     Quit from program.\n\
\n";
    if (nStringID == STRID_OUTPUTDIRNOTFOUND)
      return L"Output directory \"%s\" not found.";
    if (nStringID == STRID_OUTPUTDIREMPTY)
      return L"Output directory is empty. Create avs in input files location?";
    if (nStringID == STRID_OUTPUTDIRNOTFOUNDCONTINUE)
      return L"Output directory \"%s\" not found. Continue?";
    if (nStringID == STRID_VIRTUALDUBNOTFOUND)
      return L"VirtualDub executable not found.";
    if (nStringID == STRID_MULTIFILESFIRST)
      return L"Script contain multiple input files. Only first file will be processed. Continue?";
    if (nStringID == STRID_MULTIFILESLAST)
      return L"Script contain multiple input files. The last file setting will be imported. Continue?";
    if (nStringID == STRID_MULTIPASS)
      return L"File \"%s\" specified in script processed several times. Are you sure?";
    if (nStringID == STRID_ADDOPENSAVECOMMANDS)
      return L"Input file not detected. Add \"Open\" and \"SaveAVI\" commands?";
    if (nStringID == STRID_NOINPUTFILE)
      return L"Input file not detected.";
    if (nStringID == STRID_REMOVESUBSET)
      return L"Command \"VirtualDub.subset\" detected. Comment these lines?";
    if (nStringID == STRID_NOFILESFORPROCESS)
      return L"No files for process. List of files is empty.";
    if (nStringID == STRID_CANNOTCREATEFILE)
      return L"Cannot create file \"%s\".";
    if (nStringID == STRID_DONEFILES)
      return L"Files processeds: %d";
    if (nStringID == STRID_CREATETEMPLATE)
      return L"Create template based on script";
    if (nStringID == STRID_SCRIPTFILE)
      return L"Script:";
    if (nStringID == STRID_GENERATE)
      return L"Generate";
    if (nStringID == STRID_LISTLABEL)
      return L"List of files to batch process (drag and drop to add more):";
    if (nStringID == STRID_OUTPUTDIR)
      return L"Output directory:";
    if (nStringID == STRID_OUTPUTFILTER)
      return L"Filter output names";
    if (nStringID == STRID_FINDIT)
      return L"Find it:";
    if (nStringID == STRID_REPLACEWITH)
      return L"Replace with:";
    if (nStringID == STRID_MATCHCASE)
      return L"Match case";
    if (nStringID == STRID_REGEXP)
      return L"RegExp";
    if (nStringID == STRID_CREATE)
      return L"Create";
    if (nStringID == STRID_EXECUTE)
      return L"Execute";
    if (nStringID == STRID_SAVE)
      return L"Save...";
    if (nStringID == STRID_INPUTFILE)
      return L"Input file";
    if (nStringID == STRID_OUTPUTFILENAME)
      return L"Output file name";
    if (nStringID == STRID_EMPTY)
      return L"<EMPTY>";
    if (nStringID == STRID_TEMPLATE)
      return L"Template";
    if (nStringID == STRID_SCRIPTTYPE)
      return L"Script type";
    if (nStringID == STRID_SCRIPTVIRTUALDUB)
      return L"VirtualDub script";
    if (nStringID == STRID_SCRIPTAVS)
      return L"AVS/AVST script";
    if (nStringID == STRID_NOPARSE)
      return L"No parse entered text";
    if (nStringID == STRID_MENU_ITEMDELETE)
      return L"Delete\tDelete";
    if (nStringID == STRID_MENU_ITEMMOVEUP)
      return L"Move up\tAlt+Up";
    if (nStringID == STRID_MENU_ITEMMOVEDOWN)
      return L"Move down\tAlt+Down";
    if (nStringID == STRID_MENU_ITEMCLEAR)
      return L"Clear";
    if (nStringID == STRID_CLOSE)
      return L"Close";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}
